/*
** Module   : TransactionManager.java
** Abstract : transaction/block properties helper methods
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20050809   @22096 Created initial version, supporting
**                           transaction/block properties with full
**                           support for nested blocks, undo and retry.
** 002 GES 20050825   @22242 Added global finish processing.
** 003 ECF 20050920   @22888 Various changes to support database
**                           transaction processing. Added registration
**                           and use of ScopeableFactory interface.
**                           Reversed order of Commitable notifications.
**                           Delayed notification of opening scope until
**                           after transaction (if any) is begun.
** 004 GES 20051005   @22999 Added master commit and master finish
**                           notifications that are guaranteed to occur
**                           after all other individual commits/finishes.
** 005 ECF 20051007   @23005 More changes to support database transaction
**                           processing. Changed isTransaction()'s access
**                           to public. Removed unused, private variant of
**                           isFullTransaction() and added public, no-arg
**                           variant of this method.
** 006 GES 20051007   @23007 Added rollback notification for the updated
**                           Commitable interface AND added master
**                           rollback support.
** 007 ECF 20051104   @23234 Changes to support iterate notifications.
**                           Finalizables are now notified at each call to
**                           the iterate() method.
** 008 ECF 20051104   @23261 Update to honorError() method. Added code to
**                           report to stderr the error messages embedded
**                           in an exception chain, in the case where the
**                           error is not rethrown to a higher block.
** 009 ECF 20051108   @23277 Add deferred commit/rollback error handling.
**                           This defers throwing an exception thrown by
**                           a commit/rollback notification callback and
**                           caught during processCommit/Rollback until it
**                           is safe to do so. Logging added as well.
** 010 ECF 20051108   @23397 Added isTopLevelBlock() method. Fixed
**                           popScope(), which was popping its scope too
**                           early. Use peek instead, then pop at end of
**                           method.
** 011 ECF 20051202   @23596 Allow multiple master Commitables and
**                           Finalizables to be registered. Also extended
**                           deferred error handling to Finalizable
**                           notification processing.
** 012 GES 20060123   @24019 Moved to context-local storage and fixed a
**                           couple of deferredError bugs.
** 013 GES 20060128   @24114 Added isGlobalBlock().
** 014 GES 20060130   @24142 Added processing quit state flag for global
**                           exiting.
** 015 ECF 20060130   @24225 Added registerTopLevelFinalizable() method.
**                           Also added logging to report push/pop scope
**                           events.
** 016 GES 20060202   @24228 Added support for differentiating external
**                           blocks (external procs) from other top-level
**                           blocks (any proc/func/trigger).
** 017 GES 20060203   @24245 Changed method name makeBackup() to
**                           blockSetup().  This is done because this
**                           hook now has responsibilities to handle
**                           iterate() to compensate for usage of explicit
**                           continue (which would bypass the iterate()
**                           call).
** 018 GES 20060204   @24278 Added batch notification support.
** 019 GES 20060207   @24329 Check interrupted status in strategic hooks
**                           and throw a StopConditionException in the
**                           case where the thread has its interrupted
**                           status set.
** 020 ECF 20060206   @24344 Prevent infinite recursion loop in
**                           ContextContainer.obtain. Was caused when
**                           ScopeableFactory.createScopeable called into
**                           TransactionManager while ContextContainer was
**                           still initializing WorkArea.
** 021 GES 20060302   @24856 Added removeFinalizable() to facilitate
**                           cleanup of resources before finished().
** 022 ECF 20060309   @24997 Modified auto-remove processing for master
**                           commitables and finalizables. Auto-remove
**                           processing was too aggressive;  it must occur
**                           only for popScope for a full transaction.
** 023 GES 20060314   @25060 Moved to common logging infrastructure and
**                           added major debug logging.
** 024 GES 20060322   @25193 New rollback() signature.
** 025 ECF 20060502   @25947 Added registerNearestLoopFinalizable()
**                           method. Finds nearest enclosing block which
**                           defines a loop and registers a Finalizable to
**                           it.
** 026 GES 20060510   @26056 New state to externally query if an iteration
**                           is being called from the top or the bottom of
**                           the block.
** 027 GES 20060517   @26191 Added retry notification processing.
** 028 GES 20060518   @26274 Use an identity based map for undoables to
**                           solve problem where two keys with the same
**                           value can't be separately stored in the map.
** 029 GES 20060529   @26649 Iterate processing must trip processing
**                           rollback if there is a pending rollback
**                           since we may have gotten here via a NEXT
**                           and otherwise the rollback may not occur.
**                           Deferred rollback is also triggered before
**                           getting to the labeled block in the case
**                           where the block associated with the full
**                           transaction is enclosed (more deeply nested)
**                           in the labeled block.  The rollback is then
**                           triggered at the full transaction level.
** 030 GES 20060608   @27024 Added a method to abort further processing
**                           in a block in triggering a RETRY as part of
**                           an UNDO statement (as opposed to processing
**                           a RETRY from within a catch block).
** 031 GES 20060609   @27055 Converted deferred error processing to catch
**                           and process Exception rather than the more
**                           specific ErrorConditionException. This is
**                           required whenever we have some unexpected
**                           exception (an abnormal end).  Otherwise the
**                           deferred error handling did not trigger and
**                           the state of the TM is corrupted. This had
**                           bad effects as the finally blocks called
**                           popScope() up the stack.
** 032 ECF 20060610   @27089 Add processing of master transaction retries.
**                           Finalizables registered for master
**                           transaction processing were not receiving
**                           retry notifications.
** 033 GES 20060626   @27565 Support silent error mode in the calling
**                           method from triggerErrorInCaller().
** 034 GES 20060804   @28426 Modified block processing to ensure that
**                           the Finalizable.iterate() notifications only
**                           occur from the top of the block.
** 035 GES 20060809   @28514 Added a new concept of a lightweight block
**                           which allows the differentiation between
**                           a Progress DO block and a REPEAT/FOR block
**                           (or really any other block types in
**                           Progress).  There are new transaction level
**                           constants and helpers to test and query
**                           these levels.
** 036 GES 20060908   @29363 Retry must honor pending rollbacks for the
**                           current block.
** 037 GES 20060921   @29832 Fix to force a default retry at the same 
**                           level as any pending rollbacks if the block
**                           is not specified (if it is null). Previously
**                           only the current block was retried even when
**                           an enclosing block was the level of the
**                           rollback.  Also fixed the fact that loop
**                           protection disabling must propagate up to
**                           enclosing blocks during popScope.
** 038 ECF 20061002   @30069 Fix to rollback(). Only defer rollback if
**                           current scope is not a subtransaction or a
**                           transaction. Also prevent unnecessary retry
**                           processing from popScope().
** 039 GES 20061004   @30133 Support for nested errors (an error generated
**                           during processing of another error, while
**                           nested error mode is enabled).
** 040 GES 20061009   @30219 Honor null return from createScopeable() to
**                           allow specific factories to decide NOT to
**                           register an object for a given context.
** 041 GES 20061010   @30263 Added registerNextExternal() and deferred
**                           external procedure block registration for
**                           use by objects being constructed in an
**                           initializer/constructor but needing to be
**                           registered in the next external proc's block.
** 042 ECF 20061011   @30354 Fixed methods processFinalizables() and
**                           processRetry(). Master finalizables were not
**                           being processed in the case where no
**                           non-master finalizables were registered in
**                           the same scope.
** 043 GES 20061116   @31215 Converted all internal processing from using
**                           block labels to identify blocks, into using
**                           an index into the stack of blocks.  This is
**                           a unique value whereas the label can be
**                           duplicated in nested blocks.  In particular,
**                           there was a case where the full transaction
**                           flag was being cleared because a nested
**                           block had the same label as the one that was
**                           marked as the full transaction.
** 044 ECF 20061116   @31240 Fixed sprintf spec in WorkArea. Was using %s
**                           where %i is required, raising an exception in
**                           the Format class.
** 045 GES 20070409   @32860 When deferred errors have a condition as a
**                           cause, unpack that cause and throw that
**                           instead of the wrapper.  This negates the
**                           effect of runtime components that are not
**                           Progress aware, which wrap normal conditions
**                           in more generic exceptions.  Since the 4GL
**                           compatible runtime knows how to deal with
**                           conditions, the behavior will be correct.
** 046 NVS 20070515   @33490 Fixed ArrayIndexOutOfBoundsException in the
**                           toString() method.
** 047 GES 20071213   @36341 Disallow multiple finalizable registrations
**                           for the same object in the same block.
** 048 GES 20071213   @36352 Added deferError support to scope 
**                           notification processing.
** 049 ECF 20071214   @36377 Added protection for iteration of certain
**                           WorkArea resources. To prevent a concurrent
**                           modification exception, updates to the work
**                           area's scopeList, deferredFinalizables,
**                           masterCommit, and masterFinish collections,
**                           and to the current block's finalizables and
**                           commitables collections, are disallowed while
**                           the respective resource is being iterated.
**                           A warning is logged with a stack trace. This
**                           prevents catastrophic failure, but also
**                           signals a programming error that should be
**                           fixed.
** 050 GES 20071218   @36428 Exposed the current finalizable operation
**                           in process.
** 051 GES 20071218   @36439 Added abnormalEnd() to notify the TM about
**                           any unexpected failure, allowing any database
**                           state to be rolled back (keeping integrity
**                           of the database).
** 052 CA  20080207   @37027 Added constants for different types of blocks.
**                           Added a new pushScope(...) method which 
**                           receives the type of block being entered - 
**                           only editing blocks are currently supported. 
**                           Added a method to retrieve the type of block 
**                           which is currently active.
** 053 ECF 20080218   @37118 Fixed processRollback(). Rollback processing
**                           must occur before undo processing. Otherwise,
**                           we might be performing rollback processing on
**                           the wrong records in RecordBuffer.
** 054 ECF 20080221   @37142 Added protection code to WorkArea.toString().
**                           Also rolled back #052 (@37118), which caused
**                           regressions.
** 055 ECF 20080430   @38182 Fixed processRollback(). Reinstated #053
**                           (@37118). SVL's related fix to RecordBuffer
**                           corrects the regressions reported in #054
**                           (@37142).
** 056 GES 20080505   @38210 Moved undoables to be an array for better
**                           performance.
** 057 GES 20080612   @38756 Uncaught condition exceptions should UNDO
**                           the full transaction level as they pass
**                           through, unless the condition is QUIT (which
**                           has the behavior of committing).
** 058 GES 20080619   @38888 Massive update of infinite loop protection
**                           to fix deviations from Progress behavior.
**                           This includes ILP for NEXT (honorNext()),
**                           PAUSE support for sometimes disabling ILP
**                           (depending on runtime context), special
**                           behavior for RETRY caused by an ENDKEY,
**                           a "stutter" mode for RETRY conversion et al.
**                           Fixed some "holes" in exception processing
**                           that would allow an UNDO, RETRY cause inside
**                           a transaction but which is targeted at a
**                           block that is outside the transaction, to
**                           avoid rolling back at the full transaction
**                           level. Added the ability to deregister a
**                           list of undoables for the current scope (and
**                           nested scopes).  This is needed to allow
**                           no-undo parameters to be properly supported.
** 059 GES 20080624   @38916 Limiting registration of undoables to a
**                           single instance is being removed because
**                           it forces a compareTo which is potentially
**                           dangerous since some undoables may have
**                           errors when compared to each other (in
**                           unexpected ways).  It is also expensive so
**                           it is good to avoid it for that reason.
** 060 GES 20080625   @38960 Moved to the BlockType enum.
** 061 GES 20080627   @38992 Fix for honorNext().
** 062 ECF 20080630   @39031 Added registerNext(). Defers registration of
**                           an Undoable until the next block scope opens.
** 063 CA  20080714   @39102 Fixed deregister() memory leak: the
**                           backupUndoList was keeping the state of the 
**                           undoables after the entry of the current 
**                           block, not before.
** 064 GES 20080811   @39381 New features for UNDO and RETRY processing.
** 065 CA  20080826   @39565 Added processRollbackPending - notify the
**                           commitables that a rollback is pending to be
**                           executed by a parent block.
** 066 CA  20080815   @39448 Added honorStop(). This method will decide if 
**                           the processing of a STOP condition will be
**                           vetoed at the current scope.
** 067 GES 20080818   @39858 Added some register() signatures to provide
**                           simpler defaults. Removed the commit feature
**                           of register() since it wasn't used anywhere.
** 068 GES 20080930   @39976 Reworked undo registration to allow a cleaner
**                           interface to the caller. Now it allows
**                           varargs to minimize the number of calls.
** 069 GES 20081003   @40036 Added _isRetry().
** 070 SIY 20081002   @40015 Inferred generics. Added helper method
**                           notifyCondition().
** 071 SVL 20081024   @40219 Added findNearestExternal() function. 
** 072 SIY 20081111   @40411 Call notifyCondition() if real problem is
**                           detected in abnormalEnd().
** 073 CA  20081127   @40708 Added 'block iterated' support - method
**                           hasParentIteratedBlock will check if any
**                           of the current blocks iterated at least
**                           once.
** 074 GES 20081126   @40742 Added helpers to access the properties of
**                           a block and to manage the number of
**                           interactions the current block has had.
** 075 GES 20081210   @40866 Added support for controlling the rollup
**                           flag for the current block. Honor the rollup
**                           flag for the containing block when exiting
**                           the current block.
** 076 GES 20090115   @41131 Only wrap abend exceptions once. Modified
**                           abnormalEnd() signature to avoid duplicate
**                           notifications. In places where we re-throw
**                           a condition exception, we now ensure that
**                           abnormalEnd() is used to allow the rollback
**                           processing to happen.
** 077 GES 20090422   @41917 Converted to standard string formatting.
** 078 ECF 20090714   @43189 Reworked deferred error processing. Deferred
**                           error is now a Throwable. Fixed abnormal end
**                           processing such that an exception thrown during
**                           notification or rollback does not mask the root
**                           cause of the abnormal end.
** 079 GES 20090729   @43442 Added processValidate() method to call validate()
**                           on registered Commitable instances.
** 080 CA  20090731   @43472 Added code for off-end listener registration.
**                           When a FOR or FOR EACH block is initialized, all
**                           its associated queries must have their off-end
**                           listeners registered, to allow block termination
**                           when the query goes off-end.
** 081 CA  20090804   @43506 Added getOffEndListeners - returns the topmost 
**                           set of query off-end listeners and eventually 
**                           clears that set.
** 082 ECF 20090816   @43655 Added methods to register and deregister an
**                           OutputParameterAssigner to/from the current
**                           block.
** 083 CA  20090821   @43716 When deferring a STOP condition caused by thread
**                           interruption, it will not be logged as SEVERE
**                           (to avoid bogus stacktraces in server log).
** 084 ECF 20090826   @43785 Rolled back #008 (@23261). It is unclear why this
**                           was added and it outputs directly to stderr,
**                           seemingly unnecessarily.
** 085 GES 20090831   @43804 Save and restore error mgr silent mode flag
**                           across top-level block boundaries. Removed
**                           previous stacking support for the error mgr,
**                           which was over-stacking things.
** 086 GES 20090915   @43898 At the end of retry processing, honor deferred
**                           errors and any pending stop condition.
** 087 CA  20090917   @43931 Fix abnormal connection end. If the TM was 
**                           notified to abort processing, the next
**                           honorStopCondition call will throw a 
**                           SilentUnwindException. So, no one can interfere
**                           with the shutdown.
** 088 CA  20090922   @43979 In deferError, log SilentUnwindException's using
**                           FINER instead of SEVERE level.
** 089 GES 20090924   @44011 Added a context-specific headless flag.
** 090 SVL 20091117   @44398 Added support for the Resettable objects.
** 091 CA  20091106   @44320 Protect the rollback processing against any
**                           thread interruption incoming from a CTRL-C or a 
**                           client abend, by marking it as a critical 
**                           section.
** 092 CA  20091202   @44470 Added methods for retrieving the current nesting
**                           level and the level of the block which is started
**                           a full transaction.
** 093 GES 20100330   @44786 Added setHeadless() to allow per-context control
**                           over headless mode. This can also be done via the
**                           directory, but this interface allows a dynamic
**                           approach.
** 094 ECF 20121114          Fix for Java 7: modified pushScope to deal with
**                           varargs in overloaded register method.
** 095 ECF 20130121          Fix for Java 7: Added registerGlobal to
**                           disambiguate from overloaded register methods.
** 096 CA  20130126          Save the Block instance in a block definition
**                           field (BlockDefinition.block). 
**                           Added nearestExecutingBlock.
** 097 CA  20130209          Added dynamic EXTENT and EXTENT parameter support.
** 098 SVL 20131028          Added registerFinalizableAt, registerCommitAt, isTransactionAt.
** 099 CA  20131206          Added delayed processing of code which normally would be executed
**                           by a finalize or scopeFinished implementation. This processing will
**                           call deleted() or scopeDeleted() implementations, immediately if the
**                           current THIS-PROCEDURE is not persistent or when the procedure gets
**                           explicitly deleted, otherwise.
** 100 SVL 20140106          Added "silent" parameter to deferError function, changed function
**                           access modifier.
** 101 CA  20140109          Do not cache context-local vars, as they may get reset (in case of a
**                           state-reset appserver Agent).
** 102 CA  20140224          Finalizable and scope processing needs to be aware of persistent 
**                           procedures marked for "delete-on-exit".
** 103 SVL 20140221          findNearestTopLevel was made public.
** 104 CA  20140301          Procedure's scope is notified first for start and last for finish.
** 105 SVL 20140414          Modified isGlobalBlock in order to support agent contexts.
** 106 CA  20140513          Added a weight for the context-local var, to ensure predetermined 
**                           order during context reset.
** 107 ECF 20140813          Changed access of processValidate method to package private.
** 108 SVL 20150115          Added _isHeadless which gets "headless" attribute without initiating
**                           any contexts in TM, LT etc.
** 109 SVL 20150116          Fixed _isHeadless for conversion case.
*/

package com.goldencode.p2j.util;

import java.util.*;
import java.util.logging.*;
import com.goldencode.p2j.net.CriticalSectionManager;
import com.goldencode.p2j.net.SilentUnwindException;
import com.goldencode.p2j.persist.P2JQuery;
import com.goldencode.p2j.persist.QueryOffEndException;
import com.goldencode.p2j.persist.QueryOffEndListener;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;

/**
 * Progress compatible transaction/block properties static utility methods
 * which operate upon a context-local scoped dictionary of blocks.  This
 * class provides the following features:
 * <p>
 * <ul>
 *    <li> Maintains state for each block, handling the proper nesting of
 *         blocks.
 *    <li> Tracks the start and end of transactions and nested 
 *         sub-transactions.
 *    <li> Provides support for sending notifications of block start/end to
 *         registered instances of the {@link Scopeable} interface.
 *    <li> Provides a rollback service within a transaction or sub-transaction
 *         triggered by known error conditions or by explicit request.  See
 *         {@link Undoable} for externally driven rollback and see the
 *         {@link Commitable} for rollback notifications and custom rollback
 *         support.  See {@link #register} and {@link #registerCommit}. 
 *    <li> Provides a commit service triggered at known successful completion
 *         of a transaction or sub-transaction.  See {@link Commitable} and
 *         {@link #registerCommit} for commit notifications.
 *    <li> Provides a cleanup service for a callback on any exit from a given
 *         scope (whether that exit was normal or abnormal).  See the
 *         {@link Finalizable} interface and {@link #registerFinalizable}.
 *    <li> Provides an iteration service triggered by client code invocation
 *         of the iterate() method.  See the {@link Finalizable} interface
 *         and {@link #registerFinalizable}.
 *    <li> Provides a retry callback triggered by client code invocation
 *         of the needsRetry() method.  See the {@link Finalizable} interface
 *         and {@link #registerFinalizable}.
 *    <li> Provides a session-level cleanup service which is called after the 
 *         exit of the top-most scope. This global finish processing occurs
 *         when the last scope is popped from the stack of scopes.  For this
 *         reason, the current implementation assumes that the session is
 *         over at the moment that the top-most user-defined scope exits.
 *         See {@link #registerFinalizable} and note the use of the 
 *         <code>global</code> option.
 *    <li> Maintains state regarding retry logic to allow blocks or loops to 
 *         be retried as a result of error conditions being raised.
 *    <li> Maintains state and processing of retry and error stack unwinding.
 *    <li> Provides a session-level batch start and stop service.
 *    <li> At every block start, stop and iteration the current thread's
 *         <code>interrupted status</code> is checked and if the thread
 *         has been previously interrupted, a {@link StopConditionException}
 *         will be thrown.  See {@link #honorStopCondition}.
 * </ul>
 * <p>
 * The following is the mapping of Progress language features to the
 * corresponding feature in this class:
 * <p>
 * <pre>
 * Progress Name    Type                Method
 * ---------------  --------------      -----------------------------------
 * transaction      function            {@link #isTransactionActive}
 * retry            function            {@link #isRetry}
 * </pre>
 * <p>
 * This class minimizes the amount and complexity of client code by using
 * lists of objects to notify of specific events.  The following is a
 * summary of the event processing:
 * <p>
 * <pre>
 * Interface     Registration Method
 * ------------  ---------------------------
 * Undoable      {@link #register}
 * Commitable    {@link #registerCommit}
 * Finalizable   {@link #registerFinalizable}
 * Scopeable     {@link #registerScopeable}
 * BatchListener {@link #registerBatchListener}
 *
 * Interface     Notify Method(s)     Event
 * ------------  -------------------  ------------------------
 * Undoable      {@link #pushScope}           {@link Undoable#deepCopy}
 *               {@link #blockSetup}          {@link Undoable#deepCopy}
 *               {@link #rollback}            {@link Undoable#assign}
 *               {@link #popScope}            {@link Undoable#assign}
 * Commitable    {@link #blockSetup}          {@link Commitable#commit}
 *               {@link #popScope}            {@link Commitable#commit}
 *               {@link #rollback}            {@link Commitable#rollback}
 *               {@link #popScope}            {@link Commitable#rollback}
 * Finalizable   {@link #popScope}            {@link Finalizable#finished}
 *               {@link #blockSetup}          {@link Finalizable#iterate}
 * Scopeable     {@link #pushScope}           {@link Scopeable#scopeStart}
 *               {@link #popScope}            {@link Scopeable#scopeFinished} 
 * BatchListener {@link #batchStart}          {@link BatchListener#batchNotify}
 *               {@link #batchStop}           {@link BatchListener#batchNotify}
 * </pre>
 *
 * @author    GES
 */
public final class TransactionManager
{
   /** The block won't start a transaction or sub-transaction. */ 
   public static final int NO_TRANSACTION = 0;
   
   /** The block will start a sub-transaction if a transaction is active. */ 
   public static final int SUB_TRANSACTION = 1;
   
   /** The block will start a transaction or sub-transaction. */ 
   public static final int TRANSACTION = 2;
   
   /** No <code>Finalizable</code> processing is currently occurring. */
   public static final int OP_NONE = 0;
   
   /** <code>iterate</code>processing is currently occurring. */
   public static final int OP_ITERATE = 1;
   
   /** <code>retry</code>processing is currently occurring. */
   public static final int OP_RETRY = 2;
   
   /** <code>finished</code>processing is currently occurring. */
   public static final int OP_FINISHED = 3;
   
   /** Indicates that the block has no properties. */
   public static final int PROP_NONE = 0x00;
   
   /** ERROR property. */
   public static final int PROP_ERROR = 0x01;

   /** ENDKEY property. */
   public static final int PROP_ENDKEY = 0x02;

   /** STOP property. */
   public static final int PROP_STOP = 0x04;

   /** QUIT property. */
   public static final int PROP_QUIT = 0x08;

   /**
    * JVM-wide list of objects which create Scopeable objects to be 
    * registered in each user's session.  By design this cannot be
    * context-local.
    */
   private static final List<ScopeableFactory> scopeableFactories =
      new ArrayList<ScopeableFactory>();
   
   /** Logger (this is JVM-wide rather than being context-local). */
   private static final Logger LOG = 
      LogHelper.getLogger(TransactionManager.class.getName());
   
   /** Context-local global data area for transaction state. */
   private static ContextContainer workArea = new ContextContainer();
   
   /**
    * Generate a string representation of the given transaction level.
    *
    * @param    level
    *           The transaction level to be decoded.  Must be one of
    *           <code>NO_TRANSACTION</code>, <code>SUB_TRANSACTION</code> or
    *           <code>TRANSACTION</code>.
    *
    * @return   The human readable string.
    */
   public static String decodeLevel(int level)
   {
      String txt = "UNKNOWN";
      
      switch (level)
      {
         case NO_TRANSACTION:
            txt = "NO_TRANSACTION";
            break;
         case SUB_TRANSACTION:
            txt = "SUB_TRANSACTION";
            break;
         case TRANSACTION:
            txt = "TRANSACTION";
            break;
      }
      
      return txt;
   }
   
   /**
    * Generate a string representation of the block properties of a block.
    *
    * @param    props
    *           The block properties to decode.
    *
    * @return   The human readable string.
    */
   public static String decodeProperties(int props)
   {
      StringBuilder sb = new StringBuilder();
      
      if ((props & PROP_ERROR) != 0)
         sb.append("ERROR");
      
      if ((props & PROP_ENDKEY) != 0)
      {
         if (sb.length() > 0)
            sb.append(", ");
         
         sb.append("ENDKEY");
      }
      
      if ((props & PROP_STOP) != 0)
      {
         if (sb.length() > 0)
            sb.append(", ");
         
         sb.append("STOP");
      }
      
      if ((props & PROP_QUIT) != 0)
      {
         if (sb.length() > 0)
            sb.append(", ");
         
         sb.append("QUIT");
      }
      
      return sb.toString();
   }
   
   /**
    * Reports on the current <code>Finalizable</code> operation being
    * executed.
    *
    * @return   One of the following:
    *           <ul>
    *              <li> <code>OP_NONE</code>
    *              <li> <code>OP_ITERATE</code>
    *              <li> <code>OP_RETRY</code>
    *              <li> <code>OP_FINISHED</code>
    *           </ul>
    */
   public static int currentOperation()
   {
      return workArea.obtain().op;
   }
   
   /**
    * Register an object which will be asked to provide a
    * <code>Scopeable</code> each time a new work area context is created.
    * The resulting object is added to the work area's list of objects to
    * be notified as runtime scopes are started and finished.
    * <p>
    * These factories typically are registered during server bootstrap and
    * are not removed during the lifetime of the server process. A factory
    * should only be registered once per server process.
    *
    * @param   factory
    *          Factory to be registered.
    */
   public static void registerScopeableFactory(ScopeableFactory factory)
   {
      scopeableFactories.add(factory);
   }
   
   /**
    * Adds each element of the object array into a scoped dictionary
    * of objects requiring UNDO support.  This registry is used to backup
    * the objects at the beginning of transactions or sub-transactions.
    * <p>
    * The instances will be added in the currently opened scope.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed.
    *
    * @param    undoTargets
    *           The array of object references to be rolled back on failure.
    *           This cannot be <code>null</code>.
    */
   public static void register(Object ... undoTargets)
   {
      registerWorker(false, undoTargets);
   }
   
   /**
    * Adds each element of the object array into a scoped dictionary
    * of objects requiring UNDO support.  This registry is used to backup
    * the objects at the beginning of transactions or sub-transactions.
    * Objects are registered at the global scope.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed.
    *
    * @param    undoTargets
    *           The array of object references to be rolled back on failure.
    *           This cannot be <code>null</code>.
    */
   public static void registerGlobal(Object ... undoTargets)
   {
      registerWorker(true, undoTargets);
   }
   
   /**
    * Adds the object into a scoped dictionary of objects requiring 
    * UNDO support.  This registry is used to backup the objects at the
    * beginning of transactions or sub-transactions.
    * <p>
    * The instances will be added in the currently opened scope.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed.
    *
    * @param    undoTarget
    *           The object reference to be rolled back on failure.  This
    *           cannot be <code>null</code>.
    */
   public static void registerUndo(Undoable undoTarget)
   {
      registerUndo(workArea.obtain(), false, undoTarget);
   }
   
   /**
    * Adds each element of the object array into a scoped dictionary
    * of objects requiring UNDO support.  This registry is used to backup
    * the objects at the beginning of transactions or sub-transactions.
    * <p>
    * The instances will be added in the currently opened scope.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed.
    *
    * @param    undoTargets
    *           The array of object references to be rolled back on failure.
    *           This cannot be <code>null</code>.
    */
   public static void registerUndo(Undoable[] undoTargets)
   {
      registerUndo(false, undoTargets);
   }
   
   /**
    * Adds the object into a scoped dictionary of objects requiring 
    * UNDO support.  This registry is used to backup the objects at the
    * beginning of transactions or sub-transactions.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed.
    *
    * @param    global 
    *           Specifies the target scope. <code>false</code> specifies the
    *           current scope, and <code>true</code> specifies the global
    *           scope.
    * @param    undoTarget
    *           The object reference to be rolled back on failure.  This
    *           cannot be <code>null</code>.
    */
   public static void registerUndo(boolean global, Undoable undoTarget)
   {
      registerUndo(workArea.obtain(), global, undoTarget);
   }
   
   /**
    * Adds each element of the object array into a scoped dictionary
    * of objects requiring UNDO support.  This registry is used to backup
    * the objects at the beginning of transactions or sub-transactions.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed.
    *
    * @param    global 
    *           Specifies the target scope. <code>false</code> specifies the
    *           current scope, and <code>true</code> specifies the global
    *           scope.
    * @param    undoTargets
    *           The array of object references to be rolled back on failure.
    *           This cannot be <code>null</code>.
    */
   public static void registerUndo(boolean global, Undoable[] undoTargets)
   {
      for (Undoable next : undoTargets)
      {
         if (next != null)
            registerUndo(workArea.obtain(), global, next);
      }
   }
   
   /**
    * Remove the given extent undoables from the undoable list for the current scope and all more 
    * nested scopes. Please note that when this scope is popped off the stack of scopes, these 
    * undoables will be returned to the undoable list.
    *
    * @param    list
    *           Extent instances to temporarily remove.
    */
   public static void deregister(Undoable[][] list)
   {
      for (int i = 0; i < list.length; i++)
      {
         deregister(list[i]);
      }
   }
   
   /**
    * Remove the given undoables from the undoable list for the current
    * scope and all more nested scopes. Please note that when this scope
    * is popped off the stack of scopes, these undoables will be returned
    * to the undoable list.
    *
    * @param    list
    *           Instances to temporarily remove.
    */
   public static void deregister(Undoable[] list)
   {
      WorkArea        wa  = workArea.obtain();
      BlockDefinition blk = wa.blocks.peek();
      
      // backup the old list
      blk.backupUndoList = wa.undoList;
      
      // create a new list based on the original
      wa.undoList = new PartitionedArray<Undoable>(wa.undoList);

      // since all the deregister(...) calls are being executed within a
      // block's scope, and the undoable list is restored when this block
      // ends, we must remove the partition for the current block from the 
      // backupUndoList - this ensures the undoables are restored to the state
      // before this block has entered
      blk.backupUndoList.removePartition();

      // remove the given elements
      for (Undoable target : list)
      {
         wa.undoList.remove(target, true);
      }
      
      // refresh the local snapshot array (reads from the new wa.undoList)
      blk.undoables = createBackupSet(); 
   }
   
   /**
    * Adds the given object into the list of objects for the current block
    * which require commit notification.  This form of registration is
    * useful for objects whose natural lifetime are scoped to a different
    * block than the block in which they must be committed.  For example,
    * if a global or shared resource needs to be committed in a specific 
    * scope, then this can be used in cooperation with the normal
    * {@link #register} processing to control the commit points.
    * <p>
    * This notification will ONLY occur when there is an active transaction.
    * <p>
    * Notifications occur in LIFO order;  thus, commitables which depend upon
    * others being committed first must be registered before the objects upon
    * which they are dependent.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed.
    *
    * @param    target
    *           The object reference to be committed on success.  This
    *           cannot be <code>null</code>.
    */
   public static void registerCommit(Commitable target)
   {
      registerCommitAt(workArea.obtain().blocks.size() - 1, target);
   }

   /**
    * Adds the given object into the list of objects which require commit notification for the
    * block at the specified depth  This form of registration is useful for objects whose natural
    * lifetime are scoped to a different block than the block in which they must be committed.
    * For example, if a global or shared resource needs to be committed in a specific scope,
    * then this can be used in cooperation with the normal {@link #register} processing to control
    * the commit points.
    * <p>
    * This notification will ONLY occur when there is an active transaction.
    * <p>
    * Notifications occur in LIFO order;  thus, commitables which depend upon
    * others being committed first must be registered before the objects upon
    * which they are dependent.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed.
    *
    * @param    blockDepth
    *           Zero-based depth of the target block, starting from the outermost block.
    * @param    target
    *           The object reference to be committed on success.  This cannot be
    *           <code>null</code>.
    */
   public static void registerCommitAt(int blockDepth, Commitable target)
   {
      WorkArea wa = workArea.obtain();

      BlockDefinition blk = wa.blocks.get(blockDepth);

      // prevent ConcurrentModificationException
      if (wa.nowIterating == blk.commitables)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            log(Level.WARNING,
                "registerCommit",
                "Illegal attempt to register Commitable",
                new Throwable());
         }
         
         return;
      }
      
      // add the object to the list of commit targets
      blk.commitables.add(target);
   }
   
   /**
    * Adds the given object as the transaction level reference which will get 
    * a single, "master" commit call after all other commits have been
    * processed at the full transaction level.  This will never be called
    * for sub-transaction commits, nor will it be called if the transaction
    * is abnormally exited.
    * <p>
    * This notification will ONLY occur when there is an active transaction
    * and it only will be called at the exit of the block that corresponds
    * to the full transaction.
    * <p>
    * If the full transaction block is also an iterating block, this
    * notification will occur for each iteration of the block.
    *
    * @param    target
    *           The object reference to be committed on success.  This
    *           can be <code>null</code> if one intends to clear the current
    *           reference.
    * @param    remove
    *           If <code>true</code> automatically remove this registration
    *           after the full transaction scope has ended (whether or not
    *           the commit occurs).  If <code>false</code>, this reference
    *           will remain globally until replaced.
    */
   public static void registerTransactionCommit(Commitable target,
                                                boolean    remove)
   {
      WorkArea wa = workArea.obtain();
      wa.masterCommit.put(target, remove ? Boolean.TRUE : Boolean.FALSE);
   }
   
   /**
    * Register the given object as a handler which can veto the processing of
    * STOP condition processing at the current scope.  Handlers are invoked
    * in FIFO order.  Any handler can issue a veto, which forces the STOP
    * condition to be handled at a higher scope, if at all.
    * 
    * @param   h
    *          The stop condition veto handler.
    */
   public static void registerStopVetoHandler(StopConditionVetoHandler h)
   {
      WorkArea wa = workArea.obtain();
      wa.stopHandlers.add(h);
   }

   /**
    * Adds the given object into the list of global objects needing scope 
    * related notifications upon the entry and exit from every block. This
    * list is context-local so this registration is specific to the current
    * context.
    * <p>
    * <b>WARNING: this method should be called before the first call to
    * {@link #pushScope}. This is necessary since this list is global and
    * there is no history of which notifications have been provided. Thus,
    * if <code>pushScope</code> has been called before this method, then
    * the notifications of scope start for any scopes that have already been
    * added will never be seen by the target AND more importantly the 
    * registered target will receive end scope notifications for those
    * scopes even though there was no matching start scope notification!
    * If this method can't be called from the static initializer of this
    * class, the target must implement a facility that is safe with this
    * limitation (<code>scopeFinished</code> being called for scopes that
    * never had a matching <code>scopeStart</code>).</b>
    * <p>
    * These registrations are global and will not be removed.
    *
    * @param    target
    *           The object reference to be notified on scope start and end. 
    *           This cannot be <code>null</code>.
    */
   public static void registerScopeable(Scopeable target)
   {
      WorkArea wa = workArea.obtain();
      
      // prevent ConcurrentModificationException
      if (wa.nowIterating == wa.scopeList)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            log(Level.WARNING,
                "registerScopeable",
                "Illegal attempt to register Scopeable",
                new Throwable());
         }
         
         return;
      }
      
      // add the object to the list to receive scope notifications
      wa.scopeList.add(target);
   }

   /**
    * Register the given <code>Resettable</code> object in order to reset it
    * at specific points of block processing. 
    *
    * @param resettable
    *        <code>Resettable</code> to be registered.
    */
   public static void registerResettable(Resettable resettable)
   {
      WorkArea wa = workArea.obtain();
      wa.resettables.add(resettable);
   }
   
   /**
    * Adds the given object into the list of objects for the specified block
    * which require iterate/retry/finish notification.  This form of
    * registration is useful for objects whose natural lifetime is scoped to
    * a block and which need a hook for cleaning up resources or handling
    * other block termination processing.
    * <p>
    * This notification will occur whether or not a transaction is active.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed (after the {@link Finalizable#finished} method is called)
    * in the case of a non-global scope OR at the moment that the last
    * scope has been removed in the case of the global scope.
    * <p>
    * The given object will only be added if it is not already in the list.
    * <p>
    * Finalizables for a scope are processed in the same order as they were
    * registered.
    *
    * @param    target
    *           The object reference to be notified on any exit from the
    *           block specified.  This cannot be <code>null</code>.
    * @param    global 
    *           Specifies the target scope. <code>false</code> specifies the
    *           current scope, and <code>true</code> specifies the global
    *           scope.
    */
   public static void registerFinalizable(Finalizable target, boolean global)
   {
      int blockDepth = global ? 0 : workArea.obtain().blocks.size() - 1;
      registerFinalizableAt(blockDepth, target);
   }

   /**
    * Adds the given object into the list of objects for the specified block
    * which require iterate/retry/finish notification.  This form of
    * registration is useful for objects whose natural lifetime is scoped to
    * a block and which need a hook for cleaning up resources or handling
    * other block termination processing.
    * <p>
    * This notification will occur whether or not a transaction is active.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed (after the {@link Finalizable#finished} method is called)
    * in the case of a non-global scope OR at the moment that the last
    * scope has been removed in the case of the global scope.
    * <p>
    * The given object will only be added if it is not already in the list.
    * <p>
    * Finalizables for a scope are processed in the same order as they were
    * registered.
    *
    * @param    blockDepth
    *           zero-based depth of target block, starting from the outermost
    *           block. <code>0</code> indicates the global block.
    * @param    target
    *           The object reference to be notified on any exit from the
    *           block specified.  This cannot be <code>null</code>.
    */
   public static void registerFinalizableAt(int blockDepth, Finalizable target)
   {
      WorkArea wa = workArea.obtain();
      List<Finalizable> finishers;
      
      if (blockDepth == 0)
      {
         // 0 indicates globalBlock rather than block with the index 0 (startup block).
         finishers = wa.globalBlock.finalizables;
      }
      else
      {
         BlockDefinition blk = wa.blocks.get(blockDepth);
         finishers = blk.finalizables;
      }
      
      // prevent ConcurrentModificationException
      if (wa.nowIterating == finishers)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            log(Level.WARNING,
                "registerFinalizable",
                "Illegal attempt to register Finalizable",
                new Throwable());
         }
         
         return;
      }
      
      // add the object to the list of finalizable targets if it isn't
      // already in the list (avoids being called twice for the same event)
      if (!finishers.contains(target))
      {
         finishers.add(target);
      }
   }
   
   /**
    * Adds the given object into the list of objects which require
    * iterate/retry/finish notifications for the next block that is opened
    * which is a top-level block that corresponds to a Progress external 
    * procedure.
    * <p>
    * This form of registration is useful for objects whose natural lifetime
    * is scoped to the top-level block and which need a hook for cleaning up
    * resources or handling other block termination processing.  In 
    * particular, it is necessary for objects that are constructed during 
    * a class' constructor/initializer (before the external procedure block
    * has opened).  Such objects cannot simply register since their 
    * construction is not actually enclosed by the block.
    * <p>
    * This notification will occur whether or not a transaction is active.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed (after the {@link Finalizable#finished} method is called).
    * <p>
    * The given object will only be added if it is not already in the list.
    *
    * @param    target
    *           The object reference to be notified on any exit from the
    *           top-level external block.  This cannot be <code>null</code>.
    *
    * @throws   IllegalStateException
    *           if the top-level block cannot be identified.
    */
   public static void registerNextExternal(Finalizable target)
   {
      WorkArea wa = workArea.obtain();
      
      if (wa.deferredExtFinalizables == null)
      {
         wa.deferredExtFinalizables = new ArrayList<Finalizable>();
      }
      
      // prevent ConcurrentModificationException
      if (wa.nowIterating == wa.deferredExtFinalizables)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            log(Level.WARNING,
                "registerNextExternal",
                "Illegal attempt to register Finalizable",
                new Throwable());
         }
         
         return;
      }
      
      // add the object to the list of finalizable targets if it isn't
      // already in the list (avoids being called twice for the same event)
      if (!wa.deferredExtFinalizables.contains(target))
      {
         wa.deferredExtFinalizables.add(target);
      }
   }
   
   /**
    * Adds the given object into the list of objects which require undo
    * notifications for the next block that is opened.
    * <p>
    * This form of registration is useful for objects whose natural lifetime
    * is scoped to the top-level block.  In particular, it is necessary for
    * objects that are constructed during a class' constructor/initializer
    * (before the external procedure block has opened).  Such objects cannot
    * simply register since their construction is not actually enclosed by the
    * block.
    * <p>
    * When the deferred registration occurs, it will NOT be global and will
    * NOT be added to registered Commitables.
    * <p>
    * This notification will occur whether or not a transaction is active.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed.
    * <p>
    * The given object will only be added if it is not already in the list.
    *
    * @param    target
    *           The object reference to be registered.  This cannot be
    *           <code>null</code>.
    *
    * @throws   IllegalStateException
    *           if the top-level block cannot be identified.
    */
   public static void registerNext(Undoable target)
   {
      WorkArea wa = workArea.obtain();
      
      if (wa.deferredUndoables == null)
      {
         wa.deferredUndoables = new ArrayList<Undoable>();
      }
      
      // prevent ConcurrentModificationException
      if (wa.nowIterating == wa.deferredUndoables)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            log(Level.WARNING,
                "registerNext",
                "Illegal attempt to register Undoable",
                new Throwable());
         }
         
         return;
      }
      
      // add the object to the list of undoable targets if it isn't
      // already in the list (avoids being called twice for the same event)
      if (!wa.deferredUndoables.contains(target))
      {
         wa.deferredUndoables.add(target);
      }
   }
   
   /**
    * Adds the given object into the list of objects which require
    * iterate/retry/finish notifications for the nearest enclosing 
    * top-level block.  A top-level block is generated for Progress external 
    * procedures, internal procedures, triggers and functions.  An additional
    * parameter allows registration to be specified for the specific block
    * which corresponds with an external procedure block scope in Progress.
    * <p>
    * This form of registration is useful for objects whose natural lifetime
    * is scoped to the top-level block and which need a hook for cleaning up
    * resources or handling other block termination processing.
    * <p>
    * This notification will occur whether or not a transaction is active.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed (after the {@link Finalizable#finished} method is called).
    * <p>
    * The given object will only be added if it is not already in the list.
    * <p>
    * Finalizables for a scope are processed in the same order as they were
    * registered.
    *
    * @param    target
    *           The object reference to be notified on any exit from the
    *           top-level block.  This cannot be <code>null</code>.
    * @param    external
    *           If <code>true</code>, register with the nearest enclosing
    *           external block (external procedure), otherwise register with
    *           the nearest enclosing top-level block (which may be but is
    *           not guaranteed to be the external block).
    *
    * @throws   IllegalStateException
    *           if the top-level block cannot be identified.
    */
   public static void registerTopLevelFinalizable(Finalizable target,
                                                  boolean     external)
   {
      WorkArea       wa        = workArea.obtain();
      List<Finalizable>           finishers = null;
      Stack<BlockDefinition>          blocks    = wa.blocks;
      ListIterator<BlockDefinition>   iter      = blocks.listIterator(blocks.size());
      
      // walk backwards up the stack of block definitions 
      while (iter.hasPrevious())
      {
         BlockDefinition blk = iter.previous();
         
         // if we are looking for an external block and found one OR 
         // if we are NOT looking for an external block and found any top
         // level block
         if ((external && blk.external) || (!external && blk.topLevel))
         {
            // done!
            finishers = blk.finalizables;
            break;
         }
      }
      
      if (finishers == null)
      {
         String errmsg = (external ? "external" : "top-level");
         throw new IllegalStateException("Unable to find " + errmsg +
                                         " block");
      }
      
      // prevent ConcurrentModificationException
      if (wa.nowIterating == finishers)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            log(Level.WARNING,
                "registerTopLevelFinalizable",
                "Illegal attempt to register Finalizable",
                new Throwable());
         }
         
         return;
      }
      
      // add the object to the list of finalizable targets if it isn't
      // already in the list (avoids being called twice for the same event)
      if (!finishers.contains(target))
      {
         finishers.add(target);
      }
   }
   
   /**
    * Adds the given object into the list of objects which require
    * iterate/retry/finish notification for the nearest enclosing, looping 
    * block.
    * <p>
    * This form of registration is useful for objects whose natural lifetime
    * is scoped to the nearest enclosing, looping block and which need a hook
    * for cleaning up resources or handling other block termination
    * processing.
    * <p>
    * This notification will occur whether or not a transaction is active.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed (after the {@link Finalizable#finished} method is called).
    * <p>
    * The given object will only be added if it is not already in the list.
    *
    * @param    target
    *           The object reference to be notified on any exit from or
    *           iteration of the target block.  This cannot be
    *           <code>null</code>.
    *
    * @throws   IllegalStateException
    *           if no enclosing, looping block can be identified.
    */
   public static void registerNearestLoopFinalizable(Finalizable target)
   {
      WorkArea wa = workArea.obtain();
      
      // walk backwards up the stack of block definitions 
      Stack<BlockDefinition> blocks = wa.blocks;
      ListIterator<BlockDefinition> iter = blocks.listIterator(blocks.size());
      while (iter.hasPrevious())
      {
         BlockDefinition blk = iter.previous();
         if (blk.loop)
         {
            // prevent ConcurrentModificationException
            if (wa.nowIterating == blk.finalizables)
            {
               if (LOG.isLoggable(Level.WARNING))
               {
                  log(Level.WARNING,
                      "registerNearestLoopFinalizable",
                      "Illegal attempt to register Finalizable",
                      new Throwable());
               }
               
               return;
            }
            
            // add the object to the list of finalizable targets if it isn't
            // already in the list (avoids being called twice for the same
            // event)
            if (!blk.finalizables.contains(target))
            {
               blk.finalizables.add(target);
            }
            return;
         }
      }
      
      throw new IllegalStateException(
         "Unable to find nearest enclosing, looping block");
   }
   
   /**
    * Removes the given object from the list of objects which require
    * iterate/retry/finish notifications.  This will find the first instance
    * of the given reference by walking up each block from the current to the  
    * global block. The first instance found will be removed and the method
    * will then return. If the same instance is registered in multiple scopes,
    * it would take multiple calls to remove all instances.
    * <p>
    * After this call, no notifications will be received.
    *
    * @param    target
    *           The object reference to be removed.
    *
    * @return   <code>true</code> if the target instance was found and
    *           removed.
    */
   public static boolean removeFinalizable(Finalizable target)
   {
      if (target == null)
      {
         // nothing to do
         return false;
      }
      
      WorkArea     wa        = workArea.obtain();
      Stack<BlockDefinition>        blocks    = wa.blocks;
      ListIterator<BlockDefinition> iter      = blocks.listIterator(blocks.size());
      boolean      found     = false;
      
      // walk the stack of block definitions from the current back "up" 
      while (iter.hasPrevious() && !found)
      {
         BlockDefinition blk = iter.previous();
         if (wa.nowIterating != blk.finalizables)
         {
            found = blk.finalizables.remove(target);
         }
         else
         {
            // prevent ConcurrentModificationException
            if (LOG.isLoggable(Level.WARNING))
            {
               log(Level.WARNING,
                   "removeFinalizable",
                   "Illegal attempt to deregister Finalizable",
                   new Throwable());
            }
            
            return false;
         }
      }
      
      // if still not found, try the global block
      if (!found)
      {
         if (wa.nowIterating != wa.globalBlock.finalizables)
         {
            found = wa.globalBlock.finalizables.remove(target);
         }
         else
         {
            // prevent ConcurrentModificationException
            if (LOG.isLoggable(Level.WARNING))
            {
               log(Level.WARNING,
                   "removeFinalizable",
                   "Illegal attempt to deregister global Finalizable",
                   new Throwable());
            }
            
            return false;
         }
      }
      
      return found;
   }
   
   /**
    * Adds the given object as the transaction level reference which will get 
    * a "master" finish call after all other finish notifications have been
    * processed at the full transaction level.  This will never be called for
    * sub-transaction finish notification.
    * <p>
    * This notification will ONLY occur when there is an active transaction
    * and it only will be called at the exit of the block that corresponds
    * to the full transaction.
    * <p>
    * Multiple targets may be registered by invoking this method multiple
    * times.  However, the same object registered multiple times will receive
    * the notification only once per full transaction block exit.  If multple
    * targets are registered, they will be called in the order in which they
    * were registered.
    *
    * @param    target
    *           The object reference to be finished at full transaction block
    *           end.
    * @param    remove
    *           If <code>true</code> automatically remove this registration
    *           after the full transaction scope has ended.  If
    *           <code>false</code>, this reference will remain globally until
    *           replaced.
    */
   public static void registerTransactionFinish(Finalizable target,
                                                boolean     remove)
   {
      WorkArea wa = workArea.obtain();
      wa.masterFinish.put(target, remove ? Boolean.TRUE : Boolean.FALSE);
   }
   
   /**
    * Adds the given object into the batch notifier list.  This is a global 
    * list (it isn't associated with any block) that is triggered externally
    * by a call to {@link #batchStart} and {@link #batchStop}.  Every object
    * in the list will be notified in the order in which they were added to
    * the list.
    *
    * @param    target
    *           The object reference to be notified of any batch start and 
    *           stop.
    */
   public static void registerBatchListener(BatchListener target)
   {
      workArea.obtain().batchListeners.add(target);
   }
   
   /**
    * Reports whether the current block was just rolled back or a containing
    * block has a pending rollback. If there is no current transaction open,
    * then there cannot be a pending or current rollback.
    *
    * @return   <code>true</code> if there was or will be a rollback.
    */
   public static boolean hadRollback()
   {
      WorkArea wa = workArea.obtain();
      
      // if there is no transaction, there can't be a pending rollback or a
      // current rollback
      if (wa.transLevel != -1)
      {
         // check if a containing block has a pending rollback
         if (wa.rollbackPending)
         {
            return true;
         }
         
         BlockDefinition blk = wa.blocks.peek();
         
         // check if the current block was rolled back
         if (blk.rolledBack)
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Trigger a rollback for all registered <code>Undoable</code> objects 
    * inside the current active transaction at the scope defined by the
    * current block.  All objects that are rolled back will have their
    * values assigned based on the values backed up at the start of the
    * specified scope.  A flow control change MUST always immediately follow 
    * this method otherwise undefined results can occur.
    * <p>
    * This caller-driven control flow is required to properly duplicate the
    * semantics of Progress 4GL condition processing which always generates 
    * an implicit or explicit flow control action after every UNDO. Since the
    * flow of control change can reference a different scope than that of
    * the rollback, this can't be handled by re-throwing the exception but
    * must instead be separately driven by a deferred rollback and a caller
    * driven flow control statement such as break, continue or return. The
    * key limitation that makes this possible is the fact that the UNDO
    * (rollback) scope MUST ALWAYS BE equivalent to or more specific than
    * the scope to which flow of control will be changed.  This means that
    * one can rollback at a more deeply nested scope but change flow of
    * control to resume processing at a different (less deeply nested) scope.
    * The reverse is not possible.
    */
   public static void rollback()
   {
      rollback(null);
   }
   
   /**
    * Trigger a rollback for all registered <code>Undoable</code> objects 
    * inside the current active transaction at the scope defined
    * explicitly by the given label or implicitly if the label is
    * <code>null</code>.  All objects that are rolled back will have their
    * values assigned based on the values backed up at the start of the
    * specified scope.  If the current scope is not the target scope, the
    * given exception will be stored, the rollback will be deferred and 
    * the rollback will be processed as the scopes naturally unwind.  This
    * unwinding must be generated by a change in flow of control by the
    * calling code.  This flow control change MUST always immediately follow 
    * this method otherwise undefined results can occur.
    * <p>
    * This caller-driven control flow is required to properly duplicate the
    * semantics of Progress 4GL condition processing which always generates 
    * an implicit or explicit flow control action after every UNDO. Since the
    * flow of control change can reference a different scope than that of
    * the rollback, this can't be handled by re-throwing the exception but
    * must instead be separately driven by a deferred rollback and a caller
    * driven flow control statement such as break, continue or return. The
    * key limitation that makes this possible is the fact that the UNDO
    * (rollback) scope MUST ALWAYS BE equivalent to or more specific than
    * the scope to which flow of control will be changed.  This means that
    * one can rollback at a more deeply nested scope but change flow of
    * control to resume processing at a different (less deeply nested) scope.
    * The reverse is not possible.
    *
    * @param    label
    *           Specifies the name of the scope at which to rollback or
    *           <code>null</code> if the scope should be implicitly 
    *           determined.
    */
   public static void rollback(String label)
   {
      if (LOG.isLoggable(Level.FINE))
      {
         log(Level.FINE, "rollback", "rollback_label = " + label);
      }
      
      // find the target block and call our worker
      rollbackWorker(findStackLevel(label));
   }
      
   /**
    * Trigger a rollback for all registered <code>Undoable</code> objects 
    * inside the current active transaction at the nearest top level scope.
    * All objects that are rolled back will have their values assigned based
    * on the values backed up at the start of the specified scope.  If the
    * current scope is not the target scope, the given exception will be 
    * stored, the rollback will be deferred and the rollback will be processed
    * as the scopes naturally unwind.  This unwinding must be generated by a
    * change in flow of control by the calling code.  This flow control change 
    * MUST always immediately follow this method otherwise undefined results
    * can occur.
    * <p>
    * This caller-driven control flow is required to properly duplicate the
    * semantics of Progress 4GL condition processing which always generates 
    * an implicit or explicit flow control action after every UNDO. Since the
    * flow of control change can reference a different scope than that of
    * the rollback, this can't be handled by re-throwing the exception but
    * must instead be separately driven by a deferred rollback and a caller
    * driven flow control statement such as break, continue or return. The
    * key limitation that makes this possible is the fact that the UNDO
    * (rollback) scope MUST ALWAYS BE equivalent to or more specific than
    * the scope to which flow of control will be changed.  This means that
    * one can rollback at a more deeply nested scope but change flow of
    * control to resume processing at a different (less deeply nested) scope.
    * The reverse is not possible.
    */
   public static void rollbackTopLevel()
   {
      if (LOG.isLoggable(Level.FINE))
      {
         log(Level.FINE, "rollback nearest top level block", null);
      }
      
      // find the target block and call the worker
      rollbackWorker(findNearestTopLevel());
   }
      
   /**
    * Trigger a rollback for all registered <code>Undoable</code> objects 
    * inside the current active transaction at the scope defined by the
    * specified stack level.  All objects that are rolled back will have their
    * values assigned based on the values backed up at the start of the
    * specified scope.  If the current scope is not the target scope, the
    * given exception will be stored, the rollback will be deferred and 
    * the rollback will be processed as the scopes naturally unwind.  This
    * unwinding must be generated by a change in flow of control by the
    * calling code.  This flow control change MUST always immediately follow 
    * this method otherwise undefined results can occur.
    * <p>
    * This caller-driven control flow is required to properly duplicate the
    * semantics of Progress 4GL condition processing which always generates 
    * an implicit or explicit flow control action after every UNDO. Since the
    * flow of control change can reference a different scope than that of
    * the rollback, this can't be handled by re-throwing the exception but
    * must instead be separately driven by a deferred rollback and a caller
    * driven flow control statement such as break, continue or return. The
    * key limitation that makes this possible is the fact that the UNDO
    * (rollback) scope MUST ALWAYS BE equivalent to or more specific than
    * the scope to which flow of control will be changed.  This means that
    * one can rollback at a more deeply nested scope but change flow of
    * control to resume processing at a different (less deeply nested) scope.
    * The reverse is not possible.
    *
    * @param    level
    *           Specifies the stack level of the scope at which to rollback or
    *           -1 if the current scope should be the target.
    */
   public static void rollbackWorker(int level)
   {
      WorkArea wa = workArea.obtain();
      
      BlockDefinition blk = null;
      
      if (level == -1)
      {
         blk = wa.blocks.peek();
      }
      else
      {
         blk = wa.blocks.elementAt(level);         
      }
      
      // mark that block as being the UNDO target
      blk.isUndoLevel = true;
      
      // rollback an only occur inside an active transaction
      if (isTransaction())
      {
         // is there a pending rollback?
         if (isRollbackPending())
         {
            // this should not happen!
            String errmsg = "There is already a pending rollback.";
            throw new UnsupportedOperationException(errmsg);
         }
         else
         {
            // determine if an explicit label refers to this scope OR if
            // a lack of a label implicitly refers to this scope
            if (rollbackCurrentScope(level))
            {
               // peek at the current block def on the top of the stack
               if (level != -1)
                  blk = wa.blocks.peek();
         
               // remember that we rolled back so on exit there is no commit
               blk.rolledBack = true;
               
               // we will immediately roll back, which allows other subsequent
               // processing like retry to work properly
               processRollback(blk);
            }
            else
            {
               // the undo is destined for an enclosing scope, defer it; this
               // must be done even if the enclosing target scope is outside
               // of the transaction since otherwise, when popScope() exits
               // the transaction it will not know that there is a pending
               // rollback and the full transaction will be committed... that
               // would be a bad thing
               wa.rollbackScope   = level;
               wa.rollbackPending = true;
            }

            processResettables(true);
         }
      }
      else
      {
         // no transaction but we still need to remember that this occurred
         // so that any following retry will know the target block for an
         // implicit retry (a null label in triggerRetry/forceRetry) - note
         // that we may be setting the target scope to null right here if the
         // label parm to this method is null, that is OK
         wa.retryScope = level;

         processResettables(false);
      }
   }

   /**
    * Adds each element of the object array into a scoped dictionary
    * of objects requiring UNDO support.  This registry is used to backup
    * the objects at the beginning of transactions or sub-transactions.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed.
    *
    * @param    global 
    *           Specifies the target scope. <code>false</code> specifies the
    *           current scope, and <code>true</code> specifies the global
    *           scope.
    * @param    undoTargets
    *           The array of object references to be rolled back on failure.
    *           This cannot be <code>null</code>.
    */
   private static void registerWorker(boolean global, Object ... undoTargets)
   {
      if (undoTargets == null)
         return;
      
      WorkArea wa = workArea.obtain();
      
      for (Object next : undoTargets)
      {
         if (next instanceof Undoable)
         {
            registerUndo(wa, global, (Undoable) next);
         }
         else if (next instanceof Undoable[])
         {
            Undoable[] list = (Undoable[]) next;
            
            for (Undoable target : list)
            {
               if (target != null)
                  registerUndo(wa, global, target);
            }
         }
      }
   }
   
   /**
    * Resets all registered <code>Resettable</code> objects or makes them
    * clear their lists of changes that should be reset.
    *
    * @param clearOnly
    *        If <code>true</code> then clear the lists of changes that should
    *        be reset, otherwise perform reset of these changes.
    */
   private static void processResettables(boolean clearOnly)
   {
      WorkArea wa = workArea.obtain();

      for (Resettable resettable : wa.resettables)
      {
         resettable.resetState(clearOnly);
      }
   }
   
   /**
    * Initializes a new scope, the associated block definition data
    * structures.  A backup of objects that can be rolled back is deferred
    * until after the block opens (see {@link #blockSetup} and 
    * {@link #makeBackup}).
    * <p>
    * In the case of a block which is defined as <code>NO_TRANSACTION</code>,
    * the core data structures are initialized but no rollback support will
    * be provided.
    * <p>
    * This method will honor any pending thread interruption by throwing
    * a {@link StopConditionException} using {@link #honorStopCondition}.
    * <p>
    * This will create a block of <code>BlockType.UNKNOWN</code> and block
    * properties <code>PROP_NONE</code>.
    *
    * @param    label
    *           The text label associated with the block.  This usually
    *           should correspond with the label used in the source file
    *           to identify the block for purposes of the break and continue
    *           statements.
    * @param    level
    *           One of {@link #NO_TRANSACTION}, {@link #SUB_TRANSACTION} or
    *           {@link #TRANSACTION}.  This code defines the level of
    *           rollback and transaction support which should be provided.
    * @param    external
    *           <code>true</code> if this block is associated with the
    *           external procedure (in Progress terms). This controls the
    *           cleanup of certain resources such as temp-tables, frames
    *           and streams.
    * @param    topLevel
    *           <code>true</code> if this block is a method.
    * @param    loop
    *           Defines if this block is a loop or not.
    * @param    next
    *           <code>true</code> if this infinite loop protection should
    *           convert a retry into a next (continue), otherwise the retry
    *           will be converted into a leave (break).
    *
    * @throws   StopConditionException
    *           If the current thread's was previously interrupted.
    */
   public static void pushScope(String    label,
                                int       level,
                                boolean   external,
                                boolean   topLevel,
                                boolean   loop,
                                boolean   next)
   {
      pushScope(label, level, external, topLevel, loop, next, null);
   }

   /**
    * Initializes a new scope, the associated block definition data
    * structures.  A backup of objects that can be rolled back is deferred
    * until after the block opens (see {@link #blockSetup} and 
    * {@link #makeBackup}).
    * <p>
    * In the case of a block which is defined as <code>NO_TRANSACTION</code>,
    * the core data structures are initialized but no rollback support will
    * be provided.
    * <p>
    * This method will honor any pending thread interruption by throwing
    * a {@link StopConditionException} using {@link #honorStopCondition}.
    * <p>
    * This will create a block of <code>BlockType.UNKNOWN</code> and block
    * properties <code>PROP_NONE</code>.
    *
    * @param    label
    *           The text label associated with the block.  This usually
    *           should correspond with the label used in the source file
    *           to identify the block for purposes of the break and continue
    *           statements.
    * @param    level
    *           One of {@link #NO_TRANSACTION}, {@link #SUB_TRANSACTION} or
    *           {@link #TRANSACTION}.  This code defines the level of
    *           rollback and transaction support which should be provided.
    * @param    external
    *           <code>true</code> if this block is associated with the
    *           external procedure (in Progress terms). This controls the
    *           cleanup of certain resources such as temp-tables, frames
    *           and streams.
    * @param    topLevel
    *           <code>true</code> if this block is a method.
    * @param    loop
    *           Defines if this block is a loop or not.
    * @param    next
    *           <code>true</code> if this infinite loop protection should
    *           convert a retry into a next (continue), otherwise the retry
    *           will be converted into a leave (break).
    * @param    block
    *           The Block instance to be executed. May be null, if no Block 
    *           instance is needed (i.e. this is an internal P2J call).
    *
    * @throws   StopConditionException
    *           If the current thread's was previously interrupted.
    */
   public static void pushScope(String    label,
                                int       level,
                                boolean   external,
                                boolean   topLevel,
                                boolean   loop,
                                boolean   next,
                                Block     block)
   {
      pushScope(label,
                level,
                PROP_NONE,
                external,
                topLevel,
                loop,
                next,
                BlockType.UNKNOWN,
                block);
   }
   
   /**
    * Initializes a new scope, the associated block definition data
    * structures.  A backup of objects that can be rolled back is deferred
    * until after the block opens (see {@link #blockSetup} and 
    * {@link #makeBackup}).
    * <p>
    * In the case of a block which is defined as <code>NO_TRANSACTION</code>,
    * the core data structures are initialized but no rollback support will
    * be provided.
    * <p>
    * This method will honor any pending thread interruption by throwing
    * a {@link StopConditionException} using {@link #honorStopCondition}.
    * <p>
    * This will create a block with block properties <code>PROP_NONE</code>.
    *
    * @param    label
    *           The text label associated with the block.  This usually
    *           should correspond with the label used in the source file
    *           to identify the block for purposes of the break and continue
    *           statements.
    * @param    level
    *           One of {@link #NO_TRANSACTION}, {@link #SUB_TRANSACTION} or
    *           {@link #TRANSACTION}.  This code defines the level of
    *           rollback and transaction support which should be provided.
    * @param    external
    *           <code>true</code> if this block is associated with the
    *           external procedure (in Progress terms). This controls the
    *           cleanup of certain resources such as temp-tables, frames
    *           and streams.
    * @param    topLevel
    *           <code>true</code> if this block is a method.
    * @param    loop
    *           Defines if this block is a loop or not.
    * @param    next
    *           <code>true</code> if this infinite loop protection should
    *           convert a retry into a next (continue), otherwise the retry
    *           will be converted into a leave (break).
    * @param    blockType
    *           Specifies this block's type.
    * @param    block
    *           The Block instance to be executed.
    *
    * @throws   StopConditionException
    *           If the current thread's was previously interrupted.
    */
   public static void pushScope(String    label,
                                int       level,
                                boolean   external,
                                boolean   topLevel,
                                boolean   loop,
                                boolean   next,
                                BlockType blockType,
                                Block     block)
   {
      pushScope(label,
                level,
                PROP_NONE,
                external,
                topLevel,
                loop,
                next,
                BlockType.UNKNOWN,
                block);
   }
   
   /**
    * Initializes a new scope, the associated block definition data
    * structures.  A backup of objects that can be rolled back is deferred
    * until after the block opens (see {@link #blockSetup} and 
    * {@link #makeBackup}).
    * <p>
    * In the case of a block which is defined as <code>NO_TRANSACTION</code>,
    * the core data structures are initialized but no rollback support will
    * be provided.
    * <p>
    * This method will honor any pending thread interruption by throwing
    * a {@link StopConditionException} using {@link #honorStopCondition}
    *
    * @param    label
    *           The text label associated with the block.  This usually
    *           should correspond with the label used in the source file
    *           to identify the block for purposes of the break and continue
    *           statements.
    * @param    level
    *           One of {@link #NO_TRANSACTION}, {@link #SUB_TRANSACTION} or
    *           {@link #TRANSACTION}.  This code defines the level of
    *           rollback and transaction support which should be provided.
    * @param    props
    *           A bitfield storing block properties which will be a
    *           combination of {@link #PROP_NONE}, {@link #PROP_ERROR},
    *           {@link #PROP_ENDKEY}, {@link #PROP_STOP} or
    *           {@link #PROP_QUIT}.
    * @param    external
    *           <code>true</code> if this block is associated with the
    *           external procedure (in Progress terms). This controls the
    *           cleanup of certain resources such as temp-tables, frames
    *           and streams.
    * @param    topLevel
    *           <code>true</code> if this block is a method.
    * @param    loop
    *           Defines if this block is a loop or not.
    * @param    next
    *           <code>true</code> if this infinite loop protection should
    *           convert a retry into a next (continue), otherwise the retry
    *           will be converted into a leave (break).
    * @param    blockType
    *           Specifies this block's type.
    * @param    block
    *           The Block instance to be executed.
    *
    * @throws   StopConditionException
    *           If the current thread's was previously interrupted.
    */
   public static void pushScope(String    label,
                                int       level,
                                int       props,
                                boolean   external,
                                boolean   topLevel,
                                boolean   loop,
                                boolean   next,
                                BlockType blockType,
                                Block     block)
   {
      if (LOG.isLoggable(Level.FINE))
      {
         String spec = "new block: label = %s; type = %s; level = %s; " +
                       "external = %b; top_level = %b; loop = %b; "     +
                       "next_or_leave = %s; properties = '%s';";
                       
         Object[] parms = new Object[]
         {
            label,
            blockType.toString(),
            decodeLevel(level),
            external,
            topLevel,
            loop,
            (next ? "next" : "leave"),
            decodeProperties(props)
         };
         
         log(Level.FINE, "pushScope", String.format(spec, parms));
      }
      
      // check the interrupted status and throw the stop condition if needed
      honorStopCondition();
      
      WorkArea wa = workArea.obtain();
      
      // add the scopes
      wa.undoList.markPartition();
      
      // remember if this scope starts a new transaction rather than a
      // sub-transaction or no transaction
      boolean full = false;
      
      // start a new transaction if needed and one is not already active
      if (isTransactionLevel(level) && !isTransaction())
      {
         full = true;
      }
      
      // create the block definition
      BlockDefinition blk = new BlockDefinition(label,
                                                level,
                                                props,
                                                full,
                                                external,
                                                topLevel,
                                                loop,
                                                next,
                                                blockType,
                                                block);
      
      // add our block
      wa.blocks.push(blk);

      // add a new query off-end listeners set to the stack
      wa.offEndListeners.push(new HashSet<QueryOffEndListener>());
      
      // mark our global state as having an active transaction if needed
      if (full)
      {
         // this must be after we have added the latest block to our stack
         // in order to properly have a 0-based index
         setTransactionLevel(wa.blocks.size() - 1);
      }
      
      // the error manager needs a special notification (it can't be a
      // scopeable because it is used on both the client and the server)
      if (topLevel)
      {
         // remember if silent mode was enabled or not
         blk.suppressError = ErrorManager.isSilentError();
         
         // disable silent mode if it was enabled because silent mode on a
         // RUN statement or for a function call is something that only
         // suppresses errors during the dispatching itself and NOT during the
         // execution of the called code
         if (blk.suppressError)
         {
            ErrorManager.setSilent(false);
         }
      }
      
      // notify Scopeables that scope has opened
      processScopeNotifications(wa, blk, true);
      
      // process deferred external procedure block registration requests
      if (external && wa.deferredExtFinalizables != null)
      {
         try
         {
            wa.nowIterating = wa.deferredExtFinalizables;
            
            Iterator<Finalizable> iter =
               wa.deferredExtFinalizables.iterator();
            while (iter.hasNext())
            {
               Finalizable target = iter.next();
               registerTopLevelFinalizable(target, true);
            }
            
            // clear the list
            wa.deferredExtFinalizables = null;
         }
         finally
         {
            wa.nowIterating = null;
         }
      }
      
      // process deferred block registration requests
      if (wa.deferredUndoables != null)
      {
         try
         {
            wa.nowIterating = wa.deferredUndoables;
            
            Iterator<Undoable> iter = wa.deferredUndoables.iterator();
            while (iter.hasNext())
            {
               Undoable target = iter.next();
               register(false, new Object[] { null, target, false });
            }
            
            // clear the list
            wa.deferredUndoables = null;
         }
         finally
         {
            wa.nowIterating = null;
         }
      }
   }
   
   /**
    * Remove the current block from the stack of scopes, process any pending
    * rollbacks (if the current block is the proper target), notify any
    * objects needing commits (if the block finished successfully) and handle
    * finish processing for the block.  This method must be called from a
    * <code>finally</code> block such that this is guaranteed to be called
    * no matter how the flow of control exits the associated block (exception,
    * break, return or the flow of control naturally passing through the 
    * end of the block).
    * <p>
    * Rollback and commit processing only occur when there is an active
    * transaction.  Otherwise this processing is bypassed.  Finish processing
    * is always called.
    * <p>
    * This method implements global finish processing (as the last action
    * before returning) when the last scope is popped from the stack of
    * scopes.  For this reason, the current implementation assumes that the
    * session is over at the moment that the top-most user-defined scope
    * exits.
    * <p>
    * This method will honor any pending thread interruption by throwing
    * a {@link StopConditionException} using {@link #honorStopCondition}
    *
    * @throws   StopConditionException
    *           If the current thread's was previously interrupted.
    */
   public static void popScope()
   throws StopConditionException
   {
      if (LOG.isLoggable(Level.FINE))
      {
         log(Level.FINE, "popScope", null);
      }
      
      WorkArea wa = workArea.obtain();
      BlockDefinition blk = null;
      
      try
      {
         // get the top block def on the stack
         blk = wa.blocks.peek();
         
         // if we have no more scopes in the stack of scopes, set flag for
         // global termination
         boolean globalTerm = (wa.blocks.size() == 1);
         
         // transaction related processing
         if (isTransaction())
         {
            // is there a pending rollback?
            if (isRollbackPending())
            {
               // determine if a pending rollback refers to this scope
               if (rollbackCurrentScope(getRollbackScope()))
               {
                  processRollback(blk);
               }
               else
               {
                  processRollbackPending(blk);
               }
            }
            else
            {
               // we only commit if we haven't rolled back in this scope
               if (!blk.rolledBack)
               {
                  // OK, this is a case where we need to COMMIT!
                  processCommit(blk);
               }
            }
         }
         
         // prevent downstream code from retrying
         if (wa.retryScope != -1 && wa.retryScope == (wa.blocks.size() - 1))
         {
            wa.retryScope = -1;
         }
         
         // handle cleanup notifications
         processFinalizables(wa, blk, false);
         
         // now we are done with the scope
         processScopeNotifications(wa, blk, false);
         
         // the error manager needs a special notification (it can't be a
         // scopeable because it is used on both the client and the server)
         if (blk.topLevel && blk.suppressError)
         {
            // restore silent mode
            ErrorManager.setSilent(true);
         }      
         
         wa.undoList.removePartition();      
         
         // top level blocks must clear our transaction flag
         if (isTransactionBlock())
         {
            setTransactionLevel(-1);
         }
         
         // global termination occurs as the last thing after the last scope
         // is removed
         if (globalTerm)
         {
            processFinalizables(wa, wa.globalBlock, false);
         }
         
         // Auto-remove processing for master commitables and finalizables.
         if (isFullTransaction())
         {
            wa.autoRemoveMasterCommit();
            wa.autoRemoveMasterFinish();
         }
      }
      
      finally
      {
         // pop the top block definition
         wa.blocks.pop();
         
         // pop the query off-end listeners stack
         wa.offEndListeners.pop();

         if (blk != null)
         {
            // restore any backup Undoable registry that may exist
            if (blk.backupUndoList != null)
            {
               wa.undoList = blk.backupUndoList; 
            }
            
            // propagate any disabling of loop protection in the inner block
            // back to any enclosing block
            if (wa.blocks.size() > 0)
            {
               // save off the exiting block's loop protection values
               boolean prot   = blk.loopProtection;
               boolean pause  = blk.hadPause;
               boolean endkey = blk.endkeyRetry;
               
               BlockDefinition enclosing = null;
               
               // did the inner block have loop protection disabled
               if (!prot)
               {
                  enclosing = wa.blocks.peek();
                  enclosing.loopProtection = false;
               }
               
               // did the inner block have a pause occur
               if (pause)
               {
                  if (enclosing == null)
                     enclosing = wa.blocks.peek();
                  
                  enclosing.hadPause = true;
               }
               
               if (endkey)
               {
                  if (enclosing == null)
                     enclosing = wa.blocks.peek();
                  
                  enclosing.endkeyRetry = true;
               }
               
               if (blk.interactions > 0)
               {
                  if (enclosing == null)
                     enclosing = wa.blocks.peek();
                  
                  if (enclosing.rollup)
                     enclosing.interactions += blk.interactions; 
               }
            }
         }
      }
      
      handleDeferredError();
      
      // check the interrupted status and throw the stop condition if needed
      honorStopCondition();      
   }

   /**
    * Notify {@link ConditionListener} instances about condition.
    * 
    * @param   condition
    */
   public static void notifyCondition(BlockManager.Condition condition)
   {
      // do not pass null down
      if (condition == null)
         return;
      
      BlockDefinition blk = workArea.obtain().blocks.peek();

      if (blk.finalizables != null)
      {
         for (Finalizable target : blk.finalizables)
         {
            try
            {
               if (target instanceof ConditionListener)
               {
                  ((ConditionListener) target).condition(condition);
               }
            }
            catch (Throwable thr)
            {
               deferError("notifyCondition", thr); 
            }
         }
      }
   }

   /**
    * Notify {@link ConditionListener} instances about exception condition.
    * 
    * @param   condition
    *          The instance of {@link ConditionException}.
    */
   public static void notifyCondition(ConditionException condition)
   {
      notifyCondition(BlockManager.exceptionToCondition(condition));
   }
   
   /**
    * Check if the latest block has initialized.
    * 
    * @return   See above.
    */
   public static boolean blockInitialized()
   {
      BlockDefinition blk = workArea.obtain().blocks.peek();

      return blk.init;
   }
   
   /**
    * Provides a callback entry point for the top of the loop.  This location 
    * is required to handle the backup of undo variables via 
    * {@link #makeBackup}.  On the second and subsequent executions of this
    * method for a given block, this method handles <code>iterate</code> 
    * notifications.
    * <p>
    * It is important to note that <code>continue</code> can be called from
    * nested blocks at arbitrary depths.  This method can handle such nesting
    * at any level. 
    * <p>
    * This method will honor any pending thread interruption by throwing
    * a {@link StopConditionException} using {@link #honorStopCondition}
    *
    * @throws   StopConditionException
    *           If the current thread's was previously interrupted.
    */
   public static void blockSetup() 
   throws StopConditionException
   {
      if (LOG.isLoggable(Level.FINE))
      {
         log(Level.FINE, "blockSetup", null);
      }
      
      // read the block def at the top of the stack
      BlockDefinition blk = workArea.obtain().blocks.peek();
      
      if (blk.init)
      {
         // second or subsequent time through
         iterateWorker();
         blk.iterated = true;

         processResettables(true);
      }
      else
      {
         // first time through
         blk.init = true;
      }
      
      // now trigger the undoable backups
      backupWorker(blk);
   }
   
   /**
    * Creates or updates the backup set for all objects needing rollback 
    * support in the current block, if a transaction is active.  This is
    * split up from {@link #pushScope} and {@link #blockSetup} due to the
    * semantics of Progress in which the backup set is created or refreshed
    * after the loop control variables have been modified.
    */
   public static void makeBackup() 
   {
      backupWorker(null);
   }
   
   /**
    * Reports if there is a current transaction which is active.
    *
    * @return    <code>true</code> if a transaction is active.
    */
   public static logical isTransactionActive()
   {
      return new logical(isTransaction());
   }
   
   /**
    * Reports if the current iteration of this block or loop is a retry.
    *
    * @return    <code>true</code> if this is a retry of this block or loop
    *            (as opposed to the first time through this iteration).
    */
   public static boolean _isRetry()
   {
      // peek at the current block def on the top of the stack
      BlockDefinition blk = workArea.obtain().blocks.peek();
      
      disableLoopProtection(blk);
      
      return blk.wasRetried;
   }

   /**
    * Reports if the current iteration of this block or loop is a retry.
    *
    * @return    <code>true</code> if this is a retry of this block or loop
    *            (as opposed to the first time through this iteration).
    */
   public static logical isRetry()
   {
      return new logical(_isRetry());
   }
   
   /**
    * Records the fact that the pending RETRY is being caused by an ENDKEY
    * condition. This enables "downstream" retry processing to implement the
    * "stuttering" quirk and some behavior in honoring PAUSE as a way to
    * disable infinite loop protection.
    */
   public static void markEndkeyRetry()
   {
      // peek at the current block def on the top of the stack
      BlockDefinition blk = workArea.obtain().blocks.peek();
      blk.endkeyRetry = true;
   }
   
   /**
    * Disables infinite loop protection for the remainder of the current 
    * iteration of the innermost block.  This should be called from any
    * runtime code that blocks for user input AND from the {@link #isRetry}
    * method.
    *
    * @param    blk
    *           The block on which to operator or <code>null</code> to
    *           operate on the current (innermost) block.
    */
   public static void disableLoopProtection(BlockDefinition blk)
   {
      if (blk == null)
      {
         // peek at the current block def on the top of the stack
         blk = workArea.obtain().blocks.peek();
      }
      
      blk.loopProtection = false;
   }
   
   /**
    * This checks if any of the enclosing loops has iterated at least once.
    * 
    * @return   <code>true</code> if any of the enclosing loops has iterated
    *           at least once.
    */
   public static boolean hasParentIteratedBlock()
   {
      for (BlockDefinition blk : workArea.get().blocks)
      {
         if (blk.loop && blk.iterated)
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Determines if a NEXT operation with the given target block should
    * be honored or if the NEXT should be converted to a LEAVE. This is
    * a special form of infinite loop protection.
    * <p>
    * A NEXT is converted into a LEAVE when all of the following conditions
    * are true:
    * <p>
    * <ul>
    *   <li> The NEXT is associated with the UNDO, NEXT language statement or
    *        with an ON phrase that has UNDO, NEXT as the result of the given
    *        condition (e.g. ON ENDKEY UNDO, NEXT). It is important to note
    *        that Progress does not provide infinite loop protection for the
    *        NEXT language statement, so there is no conversion to LEAVE in
    *        that case.
    *   <li> The target block (whether implicitly determined by the lack of a
    *        label or explicitly determined by the presence of a label) must
    *        also have been the target block for the UNDO operation. It is
    *        possible for the NEXT to be targetted at a more enclosing block
    *        than the UNDO, in which case there is no conversion to LEAVE.
    *   <li> The target block type must be REPEAT, REPEAT WHILE or DO WHILE.
    *        These are the only loop types that normally convert RETRY to
    *        LEAVE instead of NEXT.  Likewise, these are the only loop types
    *        which honor infinite loop protection for NEXT by converting it
    *        to a LEAVE.
    *   <li> Infinite loop protection must not have been disabled for the
    *        target block.  In this case, any form of PAUSE is honored as
    *        one of the mechanisms to disable infinite loop protection.
    * </ul>
    * <p>
    * Since NEXT is only possible on a looping block, if the target block
    * is not a looping block then this method will return <code>false</code>.
    *
    * @param    label
    *           The label of the target block for the NEXT action.
    *
    * @return   <code>true</code> for NEXT, <code>false</code> for LEAVE.
    */
   public static boolean honorNext(String label)
   {
      BlockDefinition blk = null;
      WorkArea        wa  = workArea.obtain();
      int             lvl = -1;
      
      // check if the NEXT is targetted at the current block
      if (label == null || label.length() == 0)
      {
         // get the current block
         lvl = wa.blocks.size() - 1;
         blk = wa.blocks.peek();
      }
      else
      {
         // lookup the block target by its label
         lvl = findStackLevel(label);
         
         if (lvl >= 0)
            blk = wa.blocks.elementAt(lvl);
      }
      
      // protect from NPE, but that should not happen unless a bogus label
      // is passed
      if (blk != null)
      {
         // any non-looping block type always converts NEXT to LEAVE
         if (blk.type == BlockType.DO                  ||
             blk.type == BlockType.FOR_BLOCK           ||
             blk.type == BlockType.FOR_BLOCK_WHILE     ||
             blk.type == BlockType.FOR_BLOCK_TO        ||
             blk.type == BlockType.FOR_BLOCK_TO_WHILE  ||
             blk.type == BlockType.EXTERNAL_PROC       ||
             blk.type == BlockType.INTERNAL_PROC       ||
             blk.type == BlockType.FUNCTION            ||
             blk.type == BlockType.TRIGGER)
         {
            return false;
         }
         
         // make sure the target block is one of the looping blocks that can
         // cause NEXT to convert to LEAVE
         if (blk.type == BlockType.DO_WHILE ||
             blk.type == BlockType.REPEAT   ||
             blk.type == BlockType.REPEAT_WHILE)
         {
            // conversion only happens if there is an active UNDO operation
            // targetted at the same block
            if (blk.isUndoLevel)
            {
               boolean disable = false;
               
               // this next algorithm is needed because normally, the popping
               // of each block off the stack handles the "rollup" of the
               // ILP state, but this method is being called before that
               // rollup can occur so we simulate it here to allow us to come
               // to the same conclusion
               
               // walk up the stack to find if ILP or PAUSE has occurred
               // anywhere between the current block and the target block,
               // since disabling infinite loop protection or any PAUSE
               // causes the conversion to be avoided
               for (int i = wa.blocks.size() - 1; i >= lvl; i--)
               {
                  BlockDefinition next = wa.blocks.elementAt(i);
                                         
                  if (!next.loopProtection || next.hadPause)
                  {
                     disable = true;
                     break;
                  }
               }
               
               // if ILP is not disabled
               if (!disable)
               {
                  // convert to LEAVE
                  return false;
               }
            }
         }
      }
      
      // honor the NEXT, don't convert to LEAVE
      return true;
   }
   
   /**
    * Determines if the currently executing block should be retried.  This
    * is where infinite loop protection is implemented if it is enabled.
    *
    * @return   <code>true</code> if the block (whether or not it is a loop)
    *           should be retried.
    */
   public static boolean needsRetry()
   {  
      // peek at the current block def on the top of the stack
      BlockDefinition blk = workArea.obtain().blocks.peek();
      
      boolean noPause = true;
      
      // a PAUSE sometimes acts to disable infinite loop protection
      if (blk.hadPause)
      {
         // depending solely on the current block type
         if (blk.type == BlockType.DO_WHILE ||
             blk.type == BlockType.REPEAT   ||
             blk.type == BlockType.REPEAT_WHILE)
         {
            noPause = false;
         }
         
         // also can happen in other block types IF this is a retry caused
         // by the raising of an ENDKEY condition
         if (blk.endkeyRetry)
         {
            if (blk.type == BlockType.DO                  ||
                blk.type == BlockType.DO_TO               ||
                blk.type == BlockType.REPEAT_TO           ||
                blk.type == BlockType.FOR_BLOCK           ||
                blk.type == BlockType.FOR_BLOCK_WHILE     ||
                blk.type == BlockType.FOR_BLOCK_TO        ||
                blk.type == BlockType.FOR_BLOCK_TO_WHILE  ||
                blk.type == BlockType.FOR_LOOP            ||
                blk.type == BlockType.FOR_LOOP_WHILE      ||
                blk.type == BlockType.FOR_LOOP_TO         ||
                blk.type == BlockType.FOR_LOOP_TO_WHILE)
            {
               noPause = false;
            }
         }
      }
      
      boolean bypass = false;
      
      // in certain block types when a RETRY is caused by an ENDKEY, the
      // normal infinite loop protection mechanism is bypassed depending on
      // an internal counter; this causes a kind of "stutter" effect where
      // a RETRY on some iterations will be a real RETRY and other iterations
      // it will be converted to a NEXT
      if (blk.endkeyRetry && blk.loopProtection && noPause)
      {
         // looping blocks that modify state on each pass are the only
         // blocks that get this stuttering behavior
         if (blk.type == BlockType.DO_TO               ||
             blk.type == BlockType.REPEAT_TO           ||
             blk.type == BlockType.FOR_LOOP            ||
             blk.type == BlockType.FOR_LOOP_WHILE      ||
             blk.type == BlockType.FOR_LOOP_TO         ||
             blk.type == BlockType.FOR_LOOP_TO_WHILE)
         {
            // the ilpCount will start at -1 and then approach 1 (so the
            // first 2 passes through here will convert to NEXT), then on
            // the 3rd pass the ilpCount will back down to 0 (allowing the
            // RETRY to occur normally), then the 4th pass will increment
            // the ilpCount to 1 (converts to NEXT), then 5th pass will
            // back down to 0 (RETRY), 6th pass (NEXT), 7th pass (RETRY)...
            // causing a "stuttering" effect
            
            // maintain the "stutter" counter
            if (blk.ilpCount < 1)
            {
               // allow the retry to be converted to NEXT (all these block
               // types should have blk.next == true)
               blk.ilpCount++;
            }
            else
            {
               // cause a "stutter" (allow the RETRY to proceed as if ILP
               // was disabled)
               blk.ilpCount--;
               bypass = true;
            }
         }
      }
                                             
      // honor infinite loop protection (noPause is the way that PAUSE
      // disables infinite loop protection)
      if (blk.retry && blk.loopProtection && noPause && !bypass)
      {
         // set flag to convert RETRY into LEAVE
         if (!blk.next)
         {
            blk.pendingBreak = true;
         }
                                                         
         // unless the special flag is set above, this is equivalent to NEXT
         return false;
      }
      
      boolean result = blk.retry;
      
      // is this a retry?
      if (result)
      {
         // reset our retry state as if this was a fresh pass at the block 
         // except the wasRetried will be true this time
         blk.retry          = false;
         blk.loopProtection = true;
         blk.hadPause       = false;
         blk.endkeyRetry    = false;
         blk.pendingBreak   = false;
         blk.wasRetried     = true;
         blk.rolledBack     = false;
         blk.isUndoLevel    = false;
         
         // this is the last point before the retry pass of the loop actually
         // occurs, notify the finalizables that a retry is occurring
         processRetry(blk);
      }
      
      return result;
   }
   
   /**
    * Accesses the state of whether there is a pending break that must be 
    * honored.  This would have been generated by the infinite loop
    * protection.
    *
    * @return   <code>true</code> if a break is pending for this block.
    */
   public static boolean isBreakPending()
   {
      // peek at the current block def on the top of the stack
      BlockDefinition blk = workArea.obtain().blocks.peek();
      
      return blk.pendingBreak; 
   }
   
   /**
    * Force an error to be generated such that it will be caught in the
    * caller of this method.  The stack will be unwound and normal error 
    * processing bypassed until this method returns.  This unwinding is 
    * dependent upon catch blocks being implemented with a corresponding call
    * to {@link #honorError}.
    * <p>
    * In the case where silent error mode is enabled in the calling method,
    * the "normal" stack unwinding approach cannot be used because the 
    * exception cannot be thrown. Instead, the error manager will be updated 
    * with the error data and this method will silently return.  It is
    * critical that in such cases the calling code must immediately execute a
    * <code>return</code> statement which will maintain the proper Progress
    * behavior.  The caller can check the error manager to determine that an 
    * error occurred.
    * <p>
    * Detecting the silent error mode in the caller requires that the top
    * level block be found.  The silent error flag of the caller is cached 
    * there.
    *
    * @param    text
    *           The text to use as the error message in the exception.
    *           
    * @throws   ErrorConditionException 
    */
   public static void triggerErrorInCaller(String text)
   throws ErrorConditionException
   {
      BlockDefinition blk = nearestTopLevel();
      
      if (blk != null && blk.suppressError)
      {
         ErrorManager.addError(-1, text, true);
      }
      else
      {
         // turn on the ignore flag
         workArea.obtain().ignoreError = true;
         
         // throw the error
         throw new ErrorConditionException(text);
      }
   }
   
   /**
    * Force an error to be generated such that it will be caught in the
    * caller of this method.  The stack will be unwound and normal error 
    * processing bypassed until this method returns.  This unwinding is 
    * dependent upon catch blocks being implemented with a corresponding call
    * to {@link #honorError}.
    * <p>
    * In the case where silent error mode is enabled in the calling method,
    * the "normal" stack unwinding approach cannot be used because the 
    * exception cannot be thrown. Instead, the error manager will be updated 
    * with the error data and this method will silently return.  It is
    * critical that in such cases the calling code must immediately execute a
    * <code>return</code> statement which will maintain the proper Progress
    * behavior.  The caller can check the error manager to determine that an 
    * error occurred.
    * <p>
    * Detecting the silent error mode in the caller requires that the top
    * level block be found.  The silent error flag of the caller is cached 
    * there.
    *           
    * @throws   ErrorConditionException 
    */
   public static void triggerErrorInCaller()
   throws ErrorConditionException
   {
      triggerErrorInCaller("");
   }
   
   /**
    * Force a retry of the current block (if there is no pending rollback)
    * or retry the same block as the pending rollback, in either case
    * ensure that the flow of control will not naturally continue.
    * <p>
    * A {@link RetryUnwindException} will be thrown to unwind the stack
    * to the given scope, even if the scope is the current block.  This
    * unwinding is dependent upon catch blocks being implemented with a 
    * corresponding call to {@link #honorRetry}.
    * <p>
    * <b>This method will NEVER naturally return since an exception is ALWAYS
    * thrown.</b>  This method should generally not be called from within a
    * catch block if the intention is to retry the current block since it
    * would be outside of the block's associated <code>try</code>.  In that 
    * case, only an enclosing block could be retried.
    *
    * @throws   RetryUnwindException
    *           This method always generates this exception to ensure that
    *           natural processing will not continue.
    */
   public static void forceRetry()
   throws RetryUnwindException
   {
      forceRetry(null);
   }
   
   /**
    * Force a retry of the current block or the scope named by the given
    * label, ensuring that the flow of control will not naturally continue.
    * A {@link RetryUnwindException} will be thrown to unwind the stack
    * to the given scope, even if the scope is the current block.  This
    * unwinding is dependent upon catch blocks being implemented with a 
    * corresponding call to {@link #honorRetry}.
    * <p>
    * <b>This method will NEVER naturally return since an exception is ALWAYS
    * thrown.</b>  This method should generally not be called from within a
    * catch block if the intention is to retry the current block since it
    * would be outside of the block's associated <code>try</code>.  In that 
    * case, only an enclosing block could be retried.
    *
    * @param    label
    *           The name of the scope to retry. Specify <code>null</code> to
    *           retry the scope that has a pending rollback OR to retry
    *           the current scope if there is no pending rollback.
    *
    * @throws   RetryUnwindException
    *           This method always generates this exception to ensure that
    *           natural processing will not continue.
    */
   public static void forceRetry(String label)
   throws RetryUnwindException
   {
      WorkArea wa = workArea.obtain();
      
      int level = findStackLevel(label);
      
      if (level == -1)
      {
         if (wa.rollbackPending)
         {
            // force retry at the same level as the pending rollback
            level = wa.rollbackScope;
         }
         else
         {
            if (!isTransaction())
            {
               // in a non-transaction block we still may have an implicit
               // target specified during rollback but it would have been
               // stored in the retryScope var (it may have been set to -1
               // too, that is OK)
               level = wa.retryScope;
            }
         }
      }
      
      // is the current block the target?
      if (level == -1)
      {
         // pretend that we are doing a "normal" unwind
         level = wa.blocks.size() - 1;
      }
      
      // record the scope for the pending retry
      wa.retryScope = level;
         
      // abort current processing and jump to end of current block OR unwind
      // the stack to the specified block
      throw new RetryUnwindException("Unwinding to stack level " + level);
   }
   
   /**
    * Force a retry of the current block (if there is no pending rollback)
    * or the block targeted for a pending rollback.  If a containing scope is
    * the target of the retry, then a {@link RetryUnwindException} will be
    * thrown to unwind the stack to the given scope.  This unwinding is
    * dependent upon catch blocks being implemented with a corresponding call
    * to {@link #honorRetry}.
    * <p>
    * If the current block is not the target, the retry is deferred by
    * marking the target block and then unwinding the stack.  Otherwise, 
    * this method will return naturally.  This method is only appropriate 
    * to be called from within <code>catch</code> blocks since a natural
    * return within a <code>catch</code> block will fall through to allow
    * the retry loop to engage.
    *           
    * @throws   RetryUnwindException 
    */
   public static void triggerRetry()
   throws RetryUnwindException
   {
      triggerRetry(null);
   }
   
   /**
    * Force a retry of the current block or the scope named by the given
    * label.  If a containing scope is the target of the retry, then
    * a {@link RetryUnwindException} will be thrown to unwind the stack
    * to the given scope.  This unwinding is dependent upon catch blocks 
    * being implemented with a corresponding call to {@link #honorRetry}.
    * <p>
    * If the current block is not the target, the retry is deferred by
    * marking the target block and then unwinding the stack.  Otherwise, 
    * this method will return naturally.  This method is only appropriate 
    * to be called from within <code>catch</code> blocks since a natural
    * return within a <code>catch</code> block will fall through to allow
    * the retry loop to engage.
    *
    * @param    label
    *           The name of the scope to retry. Specify <code>null</code> to
    *           retry the scope of that has a pending rollback OR to retry
    *           the current scope if there is no pending rollback.
    *           
    * @throws   RetryUnwindException 
    */
   public static void triggerRetry(String label)
   throws RetryUnwindException
   {
      WorkArea wa = workArea.obtain();
      
      int level = findStackLevel(label);
      
      if (level == -1)
      {
         if (wa.rollbackPending)
         {
            // force retry at the same level as the pending rollback
            level = wa.rollbackScope;
         }
         else
         {
            if (!isTransaction())
            {
               // in a non-transaction block we still may have an implicit
               // target specified during rollback but it would have been
               // stored in the retryScope var (it may have been set to -1
               // too, that is OK)
               level = wa.retryScope;
            }
         }
      }
      
      // peek at the current block def on the top of the stack
      BlockDefinition blk = wa.blocks.peek();
      
      if (level == -1 || level == (wa.blocks.size() - 1))
      {
         blk.retry = true;
      }
      else
      {
         // record the scope for the pending retry
         wa.retryScope = level;
         
         // unwind the stack to that point
         throw new RetryUnwindException("Unwinding to stack level " + level);
      }
   }
   
   /**
    * Dispatch the given exception to all registered stop condition handlers
    * to see if any requires that the STOP condition be vetoed at the
    * current scope.  If any of the registered handlers indicate that the STOP
    * condition should not be honored at this level, the exception is
    * rethrown, presumably to be honored at a higher scope, or not at all.  In
    * the latter case, the program running in the current context will abort.
    * 
    * @param   exc
    *          The exception to be checked.
    * 
    * @throws  StopConditionException
    *          if the STOP condition is not honored at this scope.
    */
   public static void honorStop(StopConditionException exc)
   throws StopConditionException
   {
      WorkArea wa = workArea.obtain();
      
      for (StopConditionVetoHandler handler : wa.stopHandlers)
      {
         if (handler.vetoStop(exc))
         {
            // re-throw the exception but first it will maintain the rollback
            // state/processing
            abnormalEnd(exc, false);
         }
      }
      
      // No handlers want to veto the STOP at this scope - allow the STOP
      // to be executed
   }
   
   /**
    * Check if there is a pending retry of the current block and if so then
    * enable retry.  This is the "last half" of the processing to unwind the
    * stack up to a containing scope that is the target of a retry.  The
    * "first half" of the unwinding is implemented by throwing a
    * {@link RetryUnwindException}.  This unwinding is dependent upon catch
    * blocks being implemented to call this method.
    * <p>
    * If the current block is not the target for the pending retry, the
    * exception is re-thrown.
    *
    * @param    rue
    *           The exception to honor or ignore.
    *           
    * @throws   RetryUnwindException 
    */
   public static void honorRetry(RetryUnwindException rue)
   throws RetryUnwindException
   {
      WorkArea wa = workArea.obtain();
      
      if (wa.retryScope == (wa.blocks.size() - 1))
      {
         // peek at the current block def on the top of the stack
         BlockDefinition blk = wa.blocks.peek();
         
         blk.retry = true;
         wa.retryScope = -1;
      }
      else
      {
         // we cannot propagate the exception past the full transaction
         // level without causing a rollback(), unless there is a rollback
         // already pending in which case it will be naturally be handled
         // by popScope() when the full transaction exits, so don't force a
         // rollback again (calling this more than once must not be done
         // while there is a pending rollback); note that this is only safety
         // code since "well formed" converted 4GL code should never cause
         // this to occur since the UNDO and RETRY must always be targetted
         // at the same block AND the UNDO (a call to rollback()) should
         // always be made before the RETRY is triggered
         if (isFullTransaction() && !isRollbackPending())
         {
            rollback();
         }
         
         // we aren't there yet, re-throw the exception
         throw rue;
      }
   }
   
   /**
    * Check if the current error must unwind all the way to the caller of
    * the top-level block in this method (if the ignore flag is set).  If
    * this is the case then the error is re-thrown.  Before this occurs, if
    * the current block is the top-level block of this method, then the
    * ignore flag will be cleared.  If the ignore flag is <code>false</code> 
    * then this method returns silently.
    * <p>
    * This code also honors any nested error that was generated while another
    * error was being processed AND while nested error mode was enabled.  Any
    * such stored error will be re-thrown here.
    *
    * @param    err
    *           The exception to honor or ignore.
    *           
    * @throws   ErrorConditionException 
    */
   public static void honorError(ErrorConditionException err)
   throws ErrorConditionException
   {
      WorkArea wa = workArea.obtain();
      
      if (wa.ignoreError)
      {
         // peek at the current block def on the top of the stack
         BlockDefinition blk = wa.blocks.peek();
         
         if (blk.topLevel)
         {
            // the next catch block will handle this error (it will be in
            // the caller)
            wa.ignoreError = false;
            
            // this bypasses the commit
            blk.rolledBack = true;
         }
         
         // TODO: should this call abnormalEnd() instead so that the rollback
         // state/processing is maintained/handled?
         
         // continue to unwind
         throw err;
      }
      
      // nested error support
      if (wa.nestedError != null)
      {
         ConditionException ece = wa.nestedError; 
         
         // reset our nested error
         wa.nestedError = null;
         
         // TODO: should this call abnormalEnd() instead so that the rollback
         // state/processing is maintained/handled?
         throw ece;
      }
   }
   
   /**
    * Notifies the transaction manager that an unexpected exception or error
    * has occurred and that any transaction must be aborted. This will
    * maintain the state of the system properly, including rolling back any
    * open transaction. Once the state is properly maintained, the triggering
    * <code>Throwable</code> will be re-thrown as a 
    * <code>RuntimeException</code>.
    * <p>
    * If the exception is of type <code>ConditionException</code> or
    * <code>RetryUnwindException</code>, this method will do nothing since
    * these are used for normal flow of control purposes and are used in
    * close coordination with the transaction manager's state.  They need
    * no other special behavior to be safe.  This method is designed to
    * handle exceptions (like <code>NullPointerException</code>) which are
    * not valid in a converted 4GL program.
    *
    * @param    thr
    *           The unexpected exception or error.
    *           
    * @throws   RuntimeException 
    */
   public static void abnormalEnd(Throwable thr)
   throws RuntimeException
   {
      abnormalEnd(thr, true);
   }
   
   /**
    * Notifies the transaction manager that an unexpected exception or error
    * has occurred and that any transaction must be aborted. This will
    * maintain the state of the system properly, including rolling back any
    * open transaction. Once the state is properly maintained, the triggering
    * <code>Throwable</code> will be re-thrown as a 
    * <code>RuntimeException</code>.
    * <p>
    * If the exception is of type <code>ConditionException</code> or
    * <code>RetryUnwindException</code>, this method will do nothing since
    * these are used for normal flow of control purposes and are used in
    * close coordination with the transaction manager's state.  They need
    * no other special behavior to be safe.  This method is designed to
    * handle exceptions (like <code>NullPointerException</code>) which are
    * not valid in a converted 4GL program.
    *
    * @param    thr
    *           The unexpected exception or error.
    * @param    needsNotify
    *           <code>true</code> if condition notification must be executed
    *           (only happens if the <code>thr</code> is a condition).
    *           
    * @throws   RuntimeException 
    */
   public static void abnormalEnd(Throwable thr, boolean needsNotify)
   throws RuntimeException
   {
      if (thr instanceof ConditionException)
      {
         if (needsNotify)
            notifyCondition((ConditionException) thr);
         
         // we cannot propagate the exception past the full transaction
         // level without causing a rollback() except for QUIT which has
         // the unusual behavior of committing the transaction UNLESS one
         // specifies an ON QUIT phrase inside the transaction block or a
         // nested block which encloses the raised condition (note that in
         // that case this method will never be called so it will work fine)
         if (isFullTransaction() && !(thr instanceof QuitConditionException))
         {
            // if a rollback is already pending, then it will be naturally
            // handled by popScope() when the full transaction exits, so
            // don't force a rollback again (calling this more than once
            // must not be done while there is a pending rollback)
            if (!isRollbackPending())
            {
               rollback();
            }
         }
         
         ConditionException ce = (ConditionException) thr;
         throw ce;
      }
      else if (thr instanceof RetryUnwindException)
      {
         // this should never really get called since there should always
         // be an explicit catch block for this BEFORE any Throwable catch
         // block, but so safety purposes...
         RetryUnwindException rue = (RetryUnwindException) thr;
         honorRetry(rue);
      }
      else
      {
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, "Abnormal end; original error:", thr);
         }
         
         WorkArea wa = workArea.get();
         boolean storeDeferred = (wa.deferredError == null);
         try
         {
            if (storeDeferred)
            {
               // if there is no existing deferred error, store the current
               // error, ensuring errors thrown in this try block do not mask
               // the cause of the abnormal end, but are still logged by
               // deferError()
               wa.deferredError = thr;
            }
            notifyCondition(BlockManager.Condition.UNKNOWN);
            rollback();
         }
         finally
         {
            if (storeDeferred)
            {
               // reset deferred error to its previous state
               wa.deferredError = null;
            }
         }
         
         if (thr instanceof RuntimeException)
         {
            // the exception is already the right type so we don't have to
            // wrap it again (this avoids unnecessarily long chained log
            // entries
            throw (RuntimeException) thr;
         }
         else
         {
            // wrap the exception in a type that will propagate all the
            // way "out" as an abend
            throw new RuntimeException(thr);
         }
      }
   }
   
   /**
    * Reports if the given transaction level represents the case where the
    * block has no effect on transaction processing.
    *
    * @param    lvl
    *           The transaction level to test.
    *
    * @return   <code>true</code> if the level is <code>NO_TRANSACTION</code>.
    */
   public static boolean isNoTransactionLevel(int lvl)
   {
      return (lvl == NO_TRANSACTION);
   }
   
   /**
    * Reports if the given transaction level represents the case where the
    * block has creates a sub-transaction when a transaction is active.
    *
    * @param    lvl
    *           The transaction level to test.
    *
    * @return   <code>true</code> if the level is
    *           <code>SUB_TRANSACTION</code>.
    */
   public static boolean isSubTransactionLevel(int lvl)
   {
      return (lvl == SUB_TRANSACTION);
   }
   
   /**
    * Reports if the given transaction level represents the case where the
    * block has creates a new transaction if a transaction is not active or
    * creates a sub-transaction when a transaction is active.
    *
    * @param    lvl
    *           The transaction level to test.
    *
    * @return   <code>true</code> if the level is <code>TRANSACTION</code>.
    */
   public static boolean isTransactionLevel(int lvl)
   {
      return (lvl == TRANSACTION);
   }
   
   /**
    * Reports the transaction level of the current block.
    *
    * @return   The transaction level.
    */
   public static int currentTransactionLevel()
   {
      BlockDefinition blk = workArea.obtain().blocks.peek();
      return blk.level;
   }
   
   /**
    * Reports if the current scope defines the outermost transaction block
    * (which is the scope at which final commit processing must occur).
    *
    * @return   <code>true</code> if the current scope defines the top-level
    *           transaction.
    */
   public static boolean isFullTransaction()
   {
      BlockDefinition blk = workArea.obtain().blocks.peek();
      return blk.full;
   }
   
   /**
    * Reports if there is a current transaction which is active.
    *
    * @return    <code>true</code> if a transaction is active.
    */
   public static boolean isTransaction()
   {
      return workArea.obtain().transLevel != -1;
   }
   
   /**
    * Reports if there is a current active transaction at the specified block depth.
    *
    * @param     blockDepth
    *            Zero-based block depth starting from the outermost block at which transaction
    *            status is checked.
    *
    * @return    <code>true</code> if a transaction is active at the specified block depth.
    */
   public static boolean isTransactionAt(int blockDepth)
   {
      int trans = workArea.obtain().transLevel;
      return trans != -1 && trans <= blockDepth;
   }
   
   /**
    * Reports if the current block is a top-level block (this corresponds
    * with a method-level block).
    *
    * @return  <code>true</code> if top-level block, else <code>false</code>.
    */
   public static boolean isTopLevelBlock()
   {
      BlockDefinition blk = workArea.obtain().blocks.peek();
      return blk.topLevel;
   }
   
   /**
    * Reports if the current block is a external block (this corresponds
    * with a method-level block that was derived from a Progress external
    * procedure).  This will be <code>false</code> for blocks derived from
    * internal procedures, functions and triggers.
    *
    * @return   <code>true</code> if it is an external block, else 
    *           <code>false</code>.
    */
   public static boolean isExternalBlock()
   {
      BlockDefinition blk = workArea.obtain().blocks.peek();
      return blk.external;
   }
   
   /**
    * Reports if the current block is a global scoped block. For normal contexts:
    * <code>true</code> if the current block is the only block on the stack of block scopes; for
    * appserver agent contexts: <code>true</code> if the current block is one of two outermost
    * blocks ("appserver-agent" and "startup") on the stack of block scopes.
    *
    * @return   See above.
    */
   public static boolean isGlobalBlock()
   {
      int size = workArea.obtain().blocks.size();
      return size == 1 || (AppServerManager.isRemote() && size == 2);
   }

   /**
    * Reports if the current block has the ERROR property.
    *
    * @return   <code>true</code> if the current block has the 
    *           <code>PROP_ERROR</code> bit set in its properties bitmap.
    */
   public static boolean hasErrorProp()
   {
      return hasProperty(PROP_ERROR);
   }
   
   /**
    * Reports if the current block has the ENDKEY property.
    *
    * @return   <code>true</code> if the current block has the 
    *           <code>PROP_ENDKEY</code> bit set in its properties bitmap.
    */
   public static boolean hasEndkeyProp()
   {
      return hasProperty(PROP_ENDKEY);
   }
   
   /**
    * Reports if the current block has the STOP property.
    *
    * @return   <code>true</code> if the current block has the 
    *           <code>PROP_STOP</code> bit set in its properties bitmap.
    */
   public static boolean hasStopProp()
   {
      return hasProperty(PROP_STOP);
   }
   
   /**
    * Reports if the current block has the QUIT property.
    *
    * @return   <code>true</code> if the current block has the 
    *           <code>PROP_QUIT</code> bit set in its properties bitmap.
    */
   public static boolean hasQuitProp()
   {
      return hasProperty(PROP_QUIT);
   }
   
   /**
    * During global finish processing, this reports if the exit was caused
    * by a <code>QuitConditionException</code>.
    *
    * @return   <code>true</code> if this exit is due to a quit condition.
    */
   public static boolean isProcessingQuit()
   {
      return workArea.obtain().processingQuit;
   }
   
   /**
    * During global finish processing, this sets the flag for an exit that
    * was caused by a <code>QuitConditionException</code>.
    *
    * @param    quit
    *           <code>true</code> if this exit is due to a quit condition.
    */
   public static void setProcessingQuit(boolean quit)
   {
      workArea.obtain().processingQuit = quit;
   }
   
   /**
    * Report if this context should be treated as headless. A headless
    * session is one that has no user interaction.
    *
    * @return   <code>true</code> if this session is headless.
    */
   public static boolean isHeadless()
   {
      return workArea.obtain().headless;
   }

   /**
    * Control the headless state of this context. A headless session is one
    * that has no user interaction.
    *
    * @param    headless
    *           <code>true</code> if this session must be headless.
    */
   public static void setHeadless(boolean headless)
   {
      workArea.obtain().headless = headless;
   }
   
   /**
    * Increases the number (by 1) of interactive language statements
    * (statements which took user input except for PAUSE) that have executed
    * in the current block since the most recent start scope, iterate or
    * retry.
    */
   public static void incrementInteractions()
   {
      BlockDefinition blk = workArea.obtain().blocks.peek();
      blk.interactions++;
   }
   
   /**
    * Resets the number of interactive language statements (statements which
    * took user input except for PAUSE) that have executed in the current
    * block to 0.
    */
   public static void resetInteractions()
   {
      BlockDefinition blk = workArea.obtain().blocks.peek();
      blk.interactions = 0;
   }
   
   /**
    * Gets the number of interactive language statements (statements which
    * took user input except for PAUSE) that have executed in the current
    * block  since the most recent start scope, iterate or retry.
    *
    * @return   The number of interactions in the current block.
    */
   public static int getInteractions()
   {
      BlockDefinition blk = workArea.obtain().blocks.peek();
      return blk.interactions;
   }
   
   /**
    * Get the rollup value for the current block. If <code>true</code>,
    * then any contained blocks that exit will cause the current (containing)
    * block's interactions counter to be increased by the contained block's
    * interactions counter. Otherwise, no aggregation will occur.
    *
    * @return   The current block's rollup value.
    */
   public static boolean isRollup()
   {
      BlockDefinition blk = workArea.obtain().blocks.peek();
      return blk.rollup;
   }
   
   /**
    * Get the rollup value for the enclosing block. If <code>true</code>,
    * then the current block's exit will cause the enclosing (containing)
    * block's interactions counter to be increased by the current block's
    * interactions counter. Otherwise, no aggregation will occur.
    *
    * @return   The current block's rollup value.
    */
   public static boolean isEnclosingRollup()
   {
      WorkArea wa = workArea.obtain();
      
      int parent = wa.blocks.size() - 2;
      
      // by default, the rollup value is true
      if (parent < 0)
         return true;
      
      BlockDefinition blk = wa.blocks.elementAt(parent);
      
      return blk.rollup;
   }
   
   /**
    * Set the new rollup value for the current block. If <code>true</code>,
    * then any contained blocks that exit will cause the current (containing)
    * block's interactions counter to be increased by the contained block's
    * interactions counter. Otherwise, no aggregation will occur.
    *
    * @param    rollup
    *           The new rollup value.
    */
   public static void setRollup(boolean rollup)
   {
      BlockDefinition blk = workArea.obtain().blocks.peek();
      blk.rollup = rollup;
   }
   
   /**
    * Notify all registered {@link BatchListener} objects that a batch is 
    * starting.
    */
   public static void batchStart()
   {
      processBatchNotifications(true);
   }
      
   /**
    * Notify all registered {@link BatchListener} objects that a batch is 
    * stopping.
    */
   public static void batchStop()
   {
      processBatchNotifications(false);
   }
   
   /**
    * Tests the interrupted status of the current thread using
    * <code>Thread.interrupted</code>, if <code>true</code> then a
    * {@link StopConditionException} will be thrown. The interrupted status
    * will be <code>true</code> if a call to <code>Thread.interrupt</code>
    * was previously made.
    * <p>
    * The thread's interrupted status is cleared by this method before the
    * <code>StopConditionException</code> is thrown.
    *
    * @throws   StopConditionException
    *           If the current thread's was previously interrupted.
    * @throws   SilentUnwindException
    *           If the connection ended prematurely end the current thread 
    *           needs to stop.
    */
   public static void honorStopCondition()
   throws StopConditionException,
          SilentUnwindException
   {
      WorkArea wa = workArea.obtain();
      if (wa.abortProcessing)
      {
         throw new SilentUnwindException();
      }

      if (Thread.interrupted())
      {
         if (LOG.isLoggable(Level.FINE))
         {
            log(Level.FINE,
                "honorStopCondition",
                "STOP (honoring deferred thread interruption)!");
         }
         
         String errmsg = "Thread was asynchronously interrupted.";
         throw new StopConditionException(errmsg);
      }
   }

   /**
    * Method called when the processing needs to be terminated. This will set
    * a flag which will indicate that the processing should be ended by 
    * raising a {@link LeaveUnwindException error}, on the next 
    * {@link #honorStopCondition()} call.
    */
   public static void abortProcessing()
   {
      WorkArea wa = workArea.obtain();
      wa.abortProcessing = true;
   }
   
   /**
    * Enable/disable nested error mode. This mode allows for storing an error
    * which is generated during the processing of a different error, so that
    * during {@link #honorError} this can be re-thrown.  This duplicates some
    * obscure behavior seen in Progress.
    *
    * @param    mode
    *           <code>true</code> to enable, <code>false</code> to disable.
    */
   public static void setNestedMode(boolean mode)
   {
      workArea.obtain().nested = mode;
   }
   
   /**
    * Reports on the enable/disable state of nested error mode. This mode
    * allows for storing an error which is generated during the processing of 
    * a different error, so that during {@link #honorError} this can be
    * re-thrown.
    *
    * @return   <code>true</code> if nested error mode is enabled.
    */
   public static boolean isNestedMode()
   {
      return workArea.obtain().nested;
   }
   
   /**
    * Returns the current block type.
    *
    * @return   The type of the current block.
    */
   public static BlockType getBlockType()
   {
      BlockDefinition blk = workArea.obtain().blocks.peek();
      
      return blk.type;
   }
   
   /**
    * Returns the index of the full transaction block, or -1 if none exists.
    *
    * @return   The index of the full transaction block, or -1 if none exists.
    */
   public static int getFullTransactionBlock()
   {
      Stack<BlockDefinition> blocks = workArea.obtain().blocks;
      
      for (int i = 0; i < blocks.size(); i++)
      {
         if (blocks.get(i).full)
         {
            return i;
         }
      }
      
      return -1;
   }
   
   /**
    * Returns the current nesting level.
    *
    * @return   the current nesting level.
    */
   public static int getNestingLevel()
   {
      return workArea.obtain().blocks.size();
   }

   /**
    * Store the given exception for re-throwing during {@link #honorError} if 
    * nested error mode is enabled.  Note that only 1 error can ever be nested
    * and the last one set via this method will be the one that is honored.
    *
    * @param    err
    *           The error to store.
    */
   public static void saveNestedError(ConditionException err)
   {
      WorkArea wa = workArea.obtain();
      
      if (wa.nested)
      {
         wa.nestedError = err;
      }
   }
   
   /** 
    * Notifies this class that an interactive pause has occurred. This may
    * affect the processing of infinite loop protection depending on the
    * type of block being processed.  This state is set for the remainder of 
    * the current iteration of the innermost block.
    */
   public static void hadPause()
   {
      // peek at the current block def on the top of the stack
      BlockDefinition blk = workArea.get().blocks.peek();
      blk.hadPause = true;
   }
   
   /**
    * Search the stack of scopes and return the level of the first external
    * block found.
    *
    * @return   The stack level of the nearest enclosing external block or
    *           -1 if there is no block marked as an external block. Level
    *           number is 1-based staring form the outermost scope.
    */
   public static int findNearestExternal()
   {
      WorkArea wa = workArea.obtain();

      ListIterator<BlockDefinition> iter =
         wa.blocks.listIterator(wa.blocks.size());

      int level = wa.blocks.size() + 1;

      // walk backwards up the stack of block definitions
      while (iter.hasPrevious())
      {
         BlockDefinition blk = iter.previous();
         level --;

         if (blk.external)
         {
            return level;
         }
      }

      return -1;
   }
   
   /**
    * This method is called before the initialization of a FOR or FOR EACH 
    * block. It will clear the {@link WorkArea#offEndQueries} set and will
    * set the {@link WorkArea#offEndRegistration} flag to ON. While this flag
    * is ON, all created queries will be collected in the 
    * {@link WorkArea#offEndQueries} set.
    * <p>
    * After the block initialization is finished, the 
    * {@link #stopOffEndRegistration()} will be called.
    */
   public static void startOffEndRegistration()
   {
      WorkArea wa = workArea.obtain();

      wa.offEndRegistration = true;
      wa.offEndQueries.clear();
   }

   /**
    * This method is called after the initialization of a FOR or FOR EACH
    * block is finished. It will save all the {@link QueryOffEndListener}s
    * associated with the registered {@link WorkArea#offEndQueries queries}.
    */
   public static void stopOffEndRegistration()
   {
      WorkArea wa = workArea.obtain();
      Set<QueryOffEndListener> listeners = wa.offEndListeners.peek();
      
      for (P2JQuery query : wa.offEndQueries)
      {
         listeners.addAll(query.getOffEndListeners());
      }
      
      wa.offEndRegistration = false;
      wa.offEndQueries.clear();
   }
   
   /**
    * Register the passed query for query off-end events. The passed queries
    * are all the queries created in the init method for a FOR or FOR EACH
    * block.
    * <p>
    * These queries will be able to end the loop when the last record is 
    * reached, by overriding any user-defined END-KEY action. The action 
    * wich is executed is the default UNDO, LEAVE action for END-KEY.
    * 
    * @param  query
    *         The query which needs to be registered.
    */
   public static void registerOffEndQuery(P2JQuery query)
   {
      WorkArea wa = workArea.obtain();
      
      if (wa.offEndRegistration)
      {
         wa.offEndQueries.add(query);
      }
   }
   
   /**
    * Check if the given listener is registered for query off-end events and
    * can override the default END-KEY action (which is UNDO, LEAVE), when
    * the current block is FOR or FOR EACH.
    * 
    * @param   listener
    *          A listener to be checked if is registered for off-end events.
    * 
    * @return  <code>true</code> if the listener is registered for query 
    *          off-end event.
    */
   public static boolean isOffEndListener(QueryOffEndListener listener)
   {
      WorkArea wa = workArea.obtain();
      Set<QueryOffEndListener> listeners = wa.offEndListeners.peek();
      
      return listeners.contains(listener);         
   }
   
   /**
    * Iterate all registered off-end listeners and call the 
    * {@link QueryOffEndListener#initialize() initialize} or 
    * {@link QueryOffEndListener#finish() finish} method, depending on the 
    * flag.
    * 
    * @param   init
    *          <code>true</code> if the listeners are initialized; else, the
    *          {@link QueryOffEndListener#finish()} method is called. 
    */
   public static void processOffEndListeners(boolean init)
   {
      WorkArea wa = workArea.obtain();
      Set<QueryOffEndListener> listeners = wa.offEndListeners.peek();
      
      for (QueryOffEndListener qoel : listeners)
      {
         if (init)
         {
            qoel.initialize();
         }
         else
         {
            qoel.finish();
         }
      }
   }
   
   /**
    * Get the topmost set of off-end {@link QueryOffEndListener listeners}.
    * 
    * @param   clear
    *          <code>true</code> if the top-most off-end listeners set
    *          should be cleared.
    */
   public static Set<QueryOffEndListener> getOffEndListeners(boolean clear)
   {
      WorkArea wa = workArea.obtain();
      
      try
      {
         return new HashSet<QueryOffEndListener>(wa.offEndListeners.peek());
      }
      finally
      {
         if (clear)
         {
            wa.offEndListeners.peek().clear();
         }
      }
   }
   
   /**
    * Register the single output parameter assigner for the current block.
    * This object will be deregistered by the {@link BlockManager} when it is
    * time to process assign-backs from function/procedure output parameters
    * to the database record fields with which they are associated.
    * 
    * @param   opa
    *          An output parameter assigner.
    * 
    * @throws  IllegalStateException
    *          if a second output parameter assigner is registered for a block
    *          before the current one is deregistered.  There can be only one.
    */
   public static void
   registerOutputParameterAssigner(OutputParameterAssigner opa)
   {
      WorkArea wa = workArea.obtain();
      
      // peek at the current block def on the top of the stack
      BlockDefinition blk = wa.blocks.peek();
      
      if (blk.outputParmAssigner != null)
      {
         throw new IllegalStateException(
            "Illegal attempt to register output parameter assigner");
      }
      
      blk.outputParmAssigner = opa;
   }

   /**
    * Defer handling of an exception thrown by a commit/rollback notification
    * callback on a registered {@link Commitable}, so that iteration or pop
    * scope processing can be continued without interruption.  This ensures
    * that all <code>Commitable</code>s registered after the errant one still
    * receive their commit/rollback notifications, and that (in the case of
    * popping a scope), registered {@link Scopeable}s still receive their
    * scope finished notifications.
    * <p>
    * If no error is already pending, <code>exc</code> is stored, to be thrown
    * upon conclusion of processing the current block scope or iteration.  If
    * an error is already pending, <code>exc</code> is discarded.  In either
    * case, <code>exc</code> is logged if error-level logging is enabled.
    *
    * @param   action
    *          A brief description of the action occurring at the time the
    *          error was encountered, such as "commit" or "master rollback".
    * @param   thr
    *          The error caught during processing of a commit or rollback
    *          notification.
    *
    * @see     #handleDeferredError
    */
   public static void deferError(String action, Throwable thr)
   {
      deferError(action, thr, false);
   }

   /**
    * Defer handling of an exception thrown by a commit/rollback notification
    * callback on a registered {@link Commitable}, so that iteration or pop
    * scope processing can be continued without interruption.  This ensures
    * that all <code>Commitable</code>s registered after the errant one still
    * receive their commit/rollback notifications, and that (in the case of
    * popping a scope), registered {@link Scopeable}s still receive their
    * scope finished notifications.
    * <p>
    * If no error is already pending, <code>exc</code> is stored, to be thrown
    * upon conclusion of processing the current block scope or iteration.  If
    * an error is already pending, <code>exc</code> is discarded.  In either
    * case, <code>exc</code> is logged if error-level logging is enabled.
    *
    * @param   action
    *          A brief description of the action occurring at the time the
    *          error was encountered, such as "commit" or "master rollback".
    * @param   thr
    *          The error caught during processing of a commit or rollback
    *          notification.
    * @param   silent
    *          If <code>true</code> then exception is not logged.
    *
    * @see     #handleDeferredError
    */
   public static void deferError(String action, Throwable thr, boolean silent)
   {
      WorkArea wa = workArea.obtain();

      // when intermixing Progress-specific and non-Progress aware runtime
      // layers, it is possible for ConditionException causes to be hidden
      // within other exception types; this processing follows the causal
      // chain and un-hides any ConditionException that was obscured
      Throwable cause       = thr;
      Throwable condition   = thr;
      boolean   interrupted = false;
      boolean   silentUnwind = false;

      while (cause != null)
      {
         if (cause instanceof ConditionException)
         {
            condition = cause;
         }

         if (cause instanceof InterruptedException)
         {
            interrupted = true;
         }

         if (cause instanceof SilentUnwindException)
         {
            silentUnwind = true;
         }

         cause = cause.getCause();
      }

      if (!silent)
      {
         // if the condition is a STOP condition generated by a thread
         // interruption via CTRL+C, then we deferr the error, but change the
         // log level from SEVERE to FINE
         Level logLvl = silentUnwind ||
               (condition instanceof StopConditionException &&
                     interrupted) ? Level.FINER
               : Level.SEVERE;

         if (LOG.isLoggable(logLvl))
         {
            String msg = "Error processing " + action + "; ";
            if (wa.deferredError == null)
            {
               msg += "deferring throw until block/iteration completes";
            }
            else
            {
               msg += "deferred error already pending; new error discarded";
            }

            log(logLvl, "deferError", msg, thr);
         }
      }

      if (wa.deferredError == null)
      {
         // save the error
         wa.deferredError = condition;
      }
   }

   /**
    * Search the stack of scopes and return the level of the first top level
    * block found.
    *
    * @return   The stack level of the nearest enclosing top level block or
    *           -1 if there is no block marked as a top level block.
    */
   public static int findNearestTopLevel()
   {
      WorkArea wa = workArea.obtain();

      // walk backwards up the stack
      for (int i = wa.blocks.size() - 1; i >= 0; i--)
      {
         BlockDefinition blk = wa.blocks.elementAt(i);

         if (blk.topLevel)
         {
            return i;
         }
      }

      return -1;
   }

   /**
    * Worker that notifies all objects in the commit list that they must
    * commit (persist any state).
    * <p>
    * On the last iteration of a loop or block which ends in success (in
    * which the flow of control reaches the natural end of the block), both
    * will be called in succession.  However, other forms of exit which are
    * still considered successful (i.e. return and break which are not the
    * result of condition processing) will only execute <code>popScope</code>
    * and <code>iterate</code> will be bypassed.
    * <p>
    * It is important to note that the list of commitables in blk is
    * iterated in reverse order from that in which they are registered with
    * the <code>TransactionManager</code>.  This is done to ensure that
    * database transactions (which are registered before their corresponding
    * buffers) are always committed after all buffers are committed.
    *
    * @throws   ErrorConditionException
    *           If validation fails.
    */
   static void processValidate()
   throws ErrorConditionException
   {
      if (LOG.isLoggable(Level.FINE))
      {
         log(Level.FINE, "processValidate", null);
      }
      
      WorkArea        wa  = workArea.obtain();
      BlockDefinition blk = wa.blocks.peek();
      
      try
      {
         wa.nowIterating = blk.commitables;
         
         for (Commitable target : blk.commitables)
         {
            target.validate(blk.full);
         }
      }
      
      finally
      {
         wa.nowIterating = null;
      }
   }

   /**
    * Get the BlockDefinition instance of the nearest block of one of the
    * following types:
    * <ul>
    *    <li>{@link BlockType#EXTERNAL_PROC external procedure}</li>
    *    <li>{@link BlockType#INTERNAL_PROC internal procedure}</li>
    *    <li>{@link BlockType#FUNCTION user-defined function}</li>
    *    <li>{@link BlockType#TRIGGER trigger}</li>
    *    <li>{@link BlockType#EDITING editing}</li>
    * </ul> 
    * 
    * @return   See above.
    */
   static BlockDefinition nearestExecutingBlock()
   {
      Stack<BlockDefinition>        blocks = workArea.obtain().blocks;
      ListIterator<BlockDefinition> iter
         = blocks.listIterator(blocks.size());
      
      // walk backwards up the stack of block definitions 
      while (iter.hasPrevious())
      {
         BlockDefinition blk = iter.previous();
         
         if (blk.type == BlockType.EXTERNAL_PROC || 
             blk.type == BlockType.INTERNAL_PROC || 
             blk.type == BlockType.FUNCTION      ||
             blk.type == BlockType.TRIGGER       ||
             blk.type == BlockType.EDITING)
         {
            // done!
            return blk;
         }
      }
      
      return null;
   }

   /**
    * Deregister the output parameter assigner for the current block, if any.
    * It is expected that only the {@link BlockManager} will invoke this
    * method.
    * 
    * @return  Output parameter assigner for this block, or <code>null</code>
    *          if none was registered.
    */
   static OutputParameterAssigner deregisterOutputParameterAssigner()
   {
      WorkArea wa = workArea.obtain();
      
      // peek at the current block def on the top of the stack
      BlockDefinition blk = wa.blocks.peek();
      
      OutputParameterAssigner opa = blk.outputParmAssigner;
      
      blk.outputParmAssigner = null;
      
      return opa;
   }

   /**
    * Worker for reporting if this context should be treated as headless. A headless session is
    * one that has no user interaction.
    *
    * @return   <code>true</code> if this session is headless.
    */
   static boolean _isHeadless()
   {
      SecurityManager sm       = SecurityManager.getInstance();
      Boolean headless =  sm != null ?
                          (Boolean) sm.getToken("headless") :
                          null;

      // first check directly in the context
      if (headless != null && headless)
      {
         return true;
      }
      else
      {
         // second check in the directory
         if (Utils.getDirectoryNodeBoolean(null, "headless", false))
         {
            return true;
         }
      }

      return false;
   }

   /**
    * Adds the object into a scoped dictionary of objects requiring 
    * UNDO support.  This registry is used to backup the objects at the
    * beginning of transactions or sub-transactions.
    * <p>
    * These registrations naturally are removed when the associated scope
    * is removed.
    *
    * @param    wa
    *           The current context-specific state or <code>null</code> if
    *           that state is not available.
    * @param    global 
    *           Specifies the target scope. <code>false</code> specifies the
    *           current scope, and <code>true</code> specifies the global
    *           scope.
    * @param    undoTarget
    *           The object reference to be rolled back on failure.  This
    *           cannot be <code>null</code>.
    */
   private static void registerUndo(WorkArea wa,
                                    boolean  global,
                                    Undoable undoTarget)
   {
      if (undoTarget == null)
         return;
      
      if (wa == null)
         wa = workArea.obtain();
      
      wa.undoList.add(global, undoTarget);
   }
   
   /**
    * Search the stack of scopes and return the first top level block found.
    *
    * @return   The nearest enclosing top level block or <code>null</code> if
    *           there is no block marked as a top level block.
    */
   private static BlockDefinition nearestTopLevel()
   {
      Stack<BlockDefinition>        blocks    = workArea.obtain().blocks;
      ListIterator<BlockDefinition> iter      = blocks.listIterator(blocks.size());
      
      // walk backwards up the stack of block definitions 
      while (iter.hasPrevious())
      {
         BlockDefinition blk = iter.previous();
         
         if (blk.topLevel)
         {
            // done!
            return blk;
         }
      }
      
      return null;
   }
   
   /**
    * Notify all registered {@link BatchListener} objects that a batch is 
    * either starting or stopping.
    *
    * @param    start
    *           <code>true</code> if this is the start of a batch, otherwise
    *           <code>false</code>.
    */
   private static void processBatchNotifications(boolean start)
   {
      if (LOG.isLoggable(Level.FINE))
      {
         log(Level.FINE, 
             "processBatchNotifications",
             "start_batch = " + Boolean.toString(start));
      }
      
      Iterator<BatchListener> iter = workArea.obtain().batchListeners.iterator();
      
      while (iter.hasNext())
      {
         BatchListener dude = iter.next();
         
         dude.batchNotify(start);
      }      
   }
   
   /**
    * Allows a loop iteration event to occur in the transaction manager,
    * which triggers commit processing before the next iteration of the
    * loop begins.
    * <p>
    * This hook is called at the top of the loop, just after the loop test
    * has been passed.  If the loop test fails, then this will not be called.
    * This code will be called inside the block itself, as the first thing
    * that is done.
    * <p>
    * Note that this is only called on the second and subsequent iterations
    * of the block, no matter how the iteration occurred (via natural control
    * flow through the bottom of an iterating block OR via a Java
    * <code>continue</code> statement that caused a jump to the top of the
    * loop).  Clearing the retry logic and rollback logic is important 
    * whether or not this is a loop since non-loops can be retried (via
    * condition processing) or iterated via <code>continue</code>.
    * <p>
    * All registered {@link Finalizable Finalizables} are notified of the
    * iteration via their {@link Finalizable#iterate} callback method.
    * <p>
    * If the block is not a loop, commit/rollback processing and 
    * <code>Finalizable</code> processing will NOT be executed.
    * <p>
    * This must never be called outside of a bracketed pair of 
    * {@link #pushScope} and {@link #popScope} calls.
    * <p>
    * This method will honor any pending thread interruption by throwing
    * a {@link StopConditionException} using {@link #honorStopCondition}
    *
    * @throws   StopConditionException
    *           If the current thread's was previously interrupted.
    */
   private static void iterateWorker()
   throws StopConditionException
   {
      if (LOG.isLoggable(Level.FINE))
      {
         log(Level.FINE, "iterateWorker", null);
      }
      
      // peek at the block def on the top of the stack
      WorkArea wa = workArea.obtain();
      BlockDefinition blk = wa.blocks.peek();
      
      // transaction related processing
      if (blk.loop && isTransaction())
      {
         // is there a pending rollback or have we rolled back in this scope?
         if (isRollbackPending())
         {
            // determine if a pending rollback refers to this scope
            if (rollbackCurrentScope(getRollbackScope()))
            {
               processRollback(blk);
            }
         }
         else if (!blk.rolledBack)
         {
            // OK, this is a case where we need to COMMIT!
            processCommit(blk);
         }
      }
      
      // reset our retry/commit logic
      blk.committed      = false;
      blk.retry          = false;
      blk.loopProtection = true;
      blk.hadPause       = false;
      blk.endkeyRetry    = false;
      blk.pendingBreak   = false;
      blk.wasRetried     = false;
      blk.rolledBack     = false;
      blk.isUndoLevel    = false;
      blk.ilpCount       = 0;
      
      // notify listeners of the iteration
      if (blk.loop)
      {
         processFinalizables(wa, blk, true);
      }
      
      handleDeferredError();
      
      // check the interrupted status and throw the stop condition if needed
      honorStopCondition();            
   }
   
   /**
    * Creates or updates the backup set for all objects needing rollback 
    * support in the current block, if a transaction is active.  This is
    * split up from {@link #pushScope} and {@link #blockSetup} due to the
    * semantics of Progress in which the backup set is created or refreshed
    * after the loop control variables have been modified.
    *
    * @param    blk
    *           The block for which a backup must be made or <code>null</code>
    *           to use the current block on top of the stack.
    */
   private static void backupWorker(BlockDefinition blk) 
   {
      // read the block def at the top of the stack
      if (blk == null)
      {
         blk = workArea.obtain().blocks.peek();
      }
      
      // transactions and sub-transactions trigger backup processing
      if (isTransaction() && !isNoTransactionLevel(blk.level))
      {
         if (blk.undoables == null)
         {
            // first time backup (for this block) of the list of objects
            // needing rollback support
            blk.undoables = createBackupSet();
         }
         else
         {
            // this is an iteration, update the backup set for the next pass
            blk.undoables = updateBackupSet(blk.undoables);
         }
      }
   }
   
   /**
    * Reports if there is a current transaction which is active.
    *
    * @return    <code>true</code> if a transaction is active.
    */
   private static boolean isTransactionBlock()
   {
      WorkArea wa = workArea.obtain();
      return (wa.transLevel != -1) &&
             (wa.transLevel == (wa.blocks.size() - 1));
   }
      
   /**
    * Starts a new top-level transaction (requires that there is no current
    * transaction which is active).
    *
    * @param     level
    *            Stack level of the scope that is to be specified as the top
    *            level scope.
    */
   private static void setTransactionLevel(int level)
   {
      WorkArea wa = workArea.obtain();
      
      if (wa.transLevel != -1 && level != -1)
         throw new IllegalArgumentException("Transaction is already active!");
      
      wa.transLevel = level;
   }
   
   /**
    * Determine if the current scope represents a transaction or 
    * sub-transaction and if so, whether the given level explicitly refers
    * to this scope OR if the given level is -1, then this implicitly refers
    * to this scope.
    * <p>
    * Even if the level is not -1 (it must refer to an enclosing block), if
    * the given block is a full transaction block (the block at which the
    * transaction is scoped), then the rollback must occur at this point
    * instead of deferring the rollback until the transaction is no longer
    * active.
    *
    * @param    level
    *           The stack level of the scope to be rolled back.
    *
    * @return   <code>true</code> if the above conditions hold.
    */
   private static boolean rollbackCurrentScope(int level)
   {
      WorkArea wa = workArea.obtain();
      
      // peek at the current block def on the top of the stack
      BlockDefinition blk = wa.blocks.peek();
      
      return !isNoTransactionLevel(blk.level) &&
             (level == -1 || level == (wa.blocks.size() - 1) || blk.full);
   }
   
   /**
    * Reports if there is a rollback operation pending.
    *
    * @return    <code>true</code> if a rollback has been requested but not
    *            yet processed.
    */
   private static boolean isRollbackPending()
   {
      return workArea.obtain().rollbackPending;
   }
   
   /**
    * Returns the level specifying the rollback scope that is pending.
    *
    * @return    The level of the scope to be rolled back or -1 if the
    *            closest containing transaction or sub-transaction should
    *            be rolled back.
    */
   private static int getRollbackScope()
   {
      return workArea.obtain().rollbackScope;
   }
   
   /**
    * Find the nearest enclosing block associated with the given label.
    *
    * @param    label
    *           The block label to search for.
    *
    * @return   The block found or <code>null</code> if there is no enclosing
    *           block with that label.
    */
   @SuppressWarnings("unused")
   private static BlockDefinition findBlock(String label)
   {
      if (label != null)
      {
         WorkArea wa = workArea.obtain();
         
         int sz = wa.blocks.size(); 
         
         // walk backwards up the stack
         for (int i = sz - 1; i >= 0; i--)
         {
            BlockDefinition blk = wa.blocks.elementAt(i);
            
            if (blk.label.equals(label))
            {
               return blk;
            }
         }
      }
      
      return null;
   }
   
   /**
    * Find the stack level of the nearest enclosing block associated with the
    * given label.
    *
    * @param    label
    *           The block label to search for.
    *
    * @return   The 0-based stack level for the block that was found or -1 if
    *           there is no enclosing block with that label.
    */
   private static int findStackLevel(String label)
   {
      if (label != null)
      {
         WorkArea wa = workArea.obtain();
         
         int sz = wa.blocks.size(); 
         
         // walk backwards up the stack
         for (int i = sz - 1; i >= 0; i--)
         {
            BlockDefinition blk = wa.blocks.elementAt(i);
            
            if (blk.label.equals(label))
            {
               return i;
            }
         }
      }
      
      return -1;
   }
   
   /**
    * Notify a block that a rollback is pending to be processed by one of 
    * its parent blocks.
    *
    * @param    blk
    *           The block to be notified about a pending rollback.
    */
   private static void processRollbackPending(BlockDefinition blk)
   {
      if (LOG.isLoggable(Level.FINE))
      {
         log(Level.FINE, "processRollbackPending", null);
      }
      
      WorkArea wa = workArea.obtain();
      
      try
      {
         wa.nowIterating = blk.commitables;
         
         // notify commitables
         for (Commitable target : blk.commitables)
         {
            try
            {
               target.rollbackPending();
            }
            catch (Throwable thr)
            {
               deferError("rollbackPending", thr);
            }
         }
      }
      
      finally
      {
         wa.nowIterating = null;
      }      
   }
   
   /**
    * Worker that actually implements rollback by copying the backed up
    * values into the original reference.  Once all externally driven
    * rollback processing (the {@link Undoable} interface) is done, 
    * notifications are provided to all registered {@link Commitable}
    * instances in the current block's commitable list.
    * <p>
    * If the block is also the full transaction block, the last thing
    * done is a single master rollback call to the instance registered
    * for the {@link #registerTransactionCommit}.
    *
    * @param    blk
    *           The block to be rolled back.
    */
   private static void processRollback(BlockDefinition blk)
   {
      if (LOG.isLoggable(Level.FINE))
      {
         log(Level.FINE, "processRollback", null);
      }
      
      WorkArea wa = workArea.obtain();
      
      // mark the start of the critical section
      CriticalSectionManager.beginSection();

      try
      {
         try
         {
            wa.nowIterating = blk.commitables;
            
            // notify commitables
            for (Commitable target : blk.commitables)
            {
               try
               {
                  target.rollback(blk.full);
               }
               catch (Throwable thr)
               {
                  deferError("rollback", thr);
               }
            }
         }
         
         finally
         {
            wa.nowIterating = null;
         }
         
         // process the master rollback
         if (blk.full)
         {
            wa.notifyMasterCommit(true);
         }      
         
         // mark the work area to note that rollback has been completed, which
         // clears the deferred rollback state;  subsequent calls to the
         // isRollbackPending() method will return false.
         wa.rollbackScope   = -1;
         wa.rollbackPending = false;
   
         // process the externally driven undo
         if (blk.undoables != null)
         {
            for (int j = 0; j < blk.undoables.length; j++)
            {
               UndoablePair next = blk.undoables[j];
               next.src.assign(next.copy);
            }
         }
      }
      
      finally
      {
         // stop the critical section
         CriticalSectionManager.endSection();
      }
   }
   
   /**
    * Worker that notifies all objects in the commit list that they must
    * commit (persist any state).
    * <p>
    * On the last iteration of a loop or block which ends in success (in
    * which the flow of control reaches the natural end of the block), both
    * will be called in succession.  However, other forms of exit which are
    * still considered successful (i.e. return and break which are not the
    * result of condition processing) will only execute <code>popScope</code>
    * and <code>iterate</code> will be bypassed.
    * <p>
    * It is important to note that the list of commitables in blk is
    * iterated in reverse order from that in which they are registered with
    * the <code>TransactionManager</code>.  This is done to ensure that
    * database transactions (which are registered before their corresponding
    * buffers) are always committed after all buffers are committed.
    *
    * @param    blk
    *           The scope to be committed.
    */
   private static void processCommit(BlockDefinition blk)
   {
      if (LOG.isLoggable(Level.FINE))
      {
         log(Level.FINE, "processCommit", null);
      }
      
      if (blk.committed || blk.undoables == null)
      {
         // already done or nothing to do, just return
         return;
      }
      
      WorkArea wa = workArea.obtain();
      
      try
      {
         wa.nowIterating = blk.commitables;
         
         for (Commitable target : blk.commitables)
         {
            try
            {
               target.commit(blk.full);
            }
            catch (Throwable thr)
            {
               deferError("commit", thr);
            }
         }
      }
      
      finally
      {
         wa.nowIterating = null;
      }
      
      // process the master commit
      if (blk.full)
      {
         wa.notifyMasterCommit(false);
      }
      
      // remember that we have already handled the commit
      blk.committed = true;
   }

   /**
    * Handle a pending, deferred error, represented by an exception thrown
    * by a commit or rollback notification callback on a registered {@link
    * Commitable}, and caught within {@link #processCommit} or {@link
    * #processRollback}, respectively.  If there is currently no deferred
    * error, this method returns immediately.  Otherwise, a debug message
    * is logged, the deferred error is reset to <code>null</code>, and the
    * stored exception is thrown.  It is intended that this method is called
    * at the end of iterate or pop scope processing, after all registered
    * <code>Commitable</code>s and <code>Scopeable</code>s have been notified
    * of the state change.
    *
    * @see     #deferError
    */
   private static void handleDeferredError()
   throws RuntimeException,
          ConditionException
   {
      WorkArea wa = workArea.obtain();
      
      if (wa.deferredError == null)
      {
         // nothing to do
         return;
      }
      
      if (LOG.isLoggable(Level.SEVERE))
      {
         String msg = "Throwing deferred error (" +
                      wa.deferredError.getMessage() + ")";
         
         log(Level.SEVERE, "handleDeferredError", msg);         
      }
      
      // reset the error condition
      Throwable thr = wa.deferredError;
      wa.deferredError = null;
      
      // condition exceptions get thrown directly since they are "safe"
      if (!(thr instanceof ConditionException))
      {
         // anything else is wrapped in an unchecked exception and becomes
         // an abnormal end
         thr = new RuntimeException("Deferred error processing.", thr);
      }
      
      // honor the error
      throw (RuntimeException) thr;
   }
   
   /**
    * Worker that notifies all objects in the finalize list that the block
    * is being exited or iterated.
    *
    * @param    wa
    *           Context-local work area.
    * @param    blk
    *           The scope to be finalized.
    * @param    iterate
    *           <code>true</code> to call each target's <code>iterate</code>
    *           method or <code>false</code> to call <code>finished</code>.
    */
   private static void processFinalizables(WorkArea        wa,
                                           BlockDefinition blk,
                                           boolean         iterate)
   {
      if (LOG.isLoggable(Level.FINE))
      {
         log(Level.FINE,
             "processFinalizables",
             "iterate = " + Boolean.toString(iterate));
      }

      if (blk.finalizables != null)
      {
         BlockDefinition nearestTopLevel = nearestTopLevel();
         Object referent = null;
         boolean delayed = false;
         // can not call for start, as the procedure is not yet initialized
         if (!iterate && isTopLevelBlock() && !isGlobalBlock())
         {
            referent = ProcedureManager._thisProcedure();
            ProcedureManager.setProcessedProcedure(referent);

            if (nearestTopLevel.type == BlockType.EXTERNAL_PROC)
            {
               delayed = ProcedureManager._isPersistent(referent) && 
                         !ProcedureManager._isDelayedDelete(referent);
            }
         }

         try
         {
            wa.nowIterating = blk.finalizables;
            wa.op = iterate ? OP_ITERATE : OP_FINISHED;
            
            for (Finalizable target : blk.finalizables)
            {
               try
               {
                  if (iterate)
                  {
                     target.iterate();
                  }
                  else
                  {
                     target.finished();

                     if (delayed)
                     {
                        ProcedureManager.addFinalizable(referent, target);
                     }
                     else
                     {
                        target.deleted();
                     }
                  }
               }
               catch (Throwable thr)
               {
                  deferError((iterate ? "iterate" : "finished"), thr);
               }
            }
         }
         
         finally
         {
            wa.nowIterating = null;
            ProcedureManager.setProcessedProcedure(null);
            wa.op = OP_NONE;
         }
      }
      
      // process the master finish/iterate
      if (blk.full)
      {
         wa.notifyMasterFinish(iterate);
      }
   }
   
   /**
    * Worker that notifies all objects in the finalize list that the block
    * is being retried.  In addition, any pending rollback for this block
    * is honored here.
    *
    * @param    blk
    *           The scope to be retried.
    */
   private static void processRetry(BlockDefinition blk)
   {
      if (LOG.isLoggable(Level.FINE))
      {
         log(Level.FINE, "processRetry", null);
      }
      
      // transaction related processing
      if (isTransaction())
      {
         // is there a pending rollback or have we rolled back in this scope?
         if (isRollbackPending())
         {
            // determine if a pending rollback refers to this scope
            if (rollbackCurrentScope(getRollbackScope()))
            {
               // note that it is good that this method does NOT set the
               // blk.rolledBack flag as otherwise a commit could not occur
               // at the end of this retry iteration
               processRollback(blk);
            }
         }
      }
      
      WorkArea wa = null;
      
      if (blk.finalizables != null)
      {
         try
         {
            wa = workArea.obtain();
            wa.nowIterating = blk.finalizables;
            wa.op = OP_RETRY;
            
            for (Finalizable target : blk.finalizables)
            {
               try
               {
                  target.retry();
               }
               catch (Throwable thr)
               {
                  deferError("retry", thr);
               }
            }
         }
         
         finally
         {
            wa.nowIterating = null;
            wa.op = OP_NONE;
         }
      }
      
      // process the master retry
      if (blk.full)
      {
         if (wa == null)
         {
            wa = workArea.obtain();
         }
         wa.notifyMasterRetry();
      }
      
      handleDeferredError();
      
      // check the interrupted status and throw the stop condition if needed
      honorStopCondition();      
   }
   
   /**
    * Worker that notifies all objects in the scopeable list that a block
    * scope is being started or ended.
    *
    * @param    wa
    *           Context-local work area.
    * @param    blk
    *           The scope to be finalized.
    * @param    start
    *           If <code>true</code> generate a start scope notification,
    *           otherwise generate an end scope notification.
    */
   private static void processScopeNotifications(WorkArea wa, BlockDefinition blk, boolean start)
   {
      if (LOG.isLoggable(Level.FINE))
      {
         log(Level.FINE,
             "processScopeNotifications",
             "start = " + Boolean.toString(start));
      }
      
      try
      {
         BlockDefinition nearestTopLevel = nearestTopLevel();
         Object referent = null;
         boolean delayed = false;
         // can not call for start, as the procedure is not yet initialized
         if (!start && isTopLevelBlock() && !isGlobalBlock())
         {
            referent = ProcedureManager._thisProcedure();
            ProcedureManager.setProcessedProcedure(referent);

            if (nearestTopLevel.type == BlockType.EXTERNAL_PROC)
            {
               delayed = ProcedureManager._isPersistent(referent) && 
                         !ProcedureManager._isDelayedDelete(referent);
            }
         }

         if (start)
         {
            // procedure scopes need to be opened first
            ProcedureManager.getScopeable().scopeStart();
         }
         
         wa.nowIterating = wa.scopeList;
         for (Scopeable target : wa.scopeList)
         {
            try
            {
               if (start)
               {
                  target.scopeStart();
               }
               else
               {
                  target.scopeFinished();

                  if (delayed)
                  {
                     ProcedureManager.addScopeable(referent, target);
                  }
                  else
                  {
                     target.scopeDeleted();
                  }
               }
            }
            
            catch (Throwable thr)
            {
               deferError((start ? "scopeStart" : "scopeFinished"), thr);
            }
         }
         
         if (!start)
         {
            // procedure scopes need to be finished last
            ProcedureManager.getScopeable().scopeFinished();
         }
      }
      
      finally
      {
         wa.nowIterating = null;
         ProcedureManager.setProcessedProcedure(null);
      }
   }
   
   /**
    * Walk through the registry of all objects needing to be backed up
    * at the beginning of a transaction or sub-transaction scope, for each
    * one make a backup and store the original reference and the backup
    * copy in a map for later rollback if needed. 
    *
    * @return   All objects and their associated backup copies. 
    */
   private static UndoablePair[] createBackupSet()
   {
      WorkArea wa = workArea.obtain();
      
      Undoable[]     var  = wa.undoList.toArray(new Undoable[0]);
      UndoablePair[] list = new UndoablePair[var.length];
      
      for (int j = 0; j < var.length; j++)
      {
         list[j] = new UndoablePair(var[j], var[j].deepCopy());
      }
      
      return list;
   }
   
   /**
    * Walk through the current backup set and refresh the stored value for
    * each one.  This processing corresponds to the end of a sub-transaction
    * (see {@link #blockSetup}). 
    *
    * @return   All objects and their updated backup copies. 
    */
   private static UndoablePair[] updateBackupSet(UndoablePair[] backup)
   {
      for (int j = 0; j < backup.length; j++)
      {
         backup[j].copy = backup[j].src.deepCopy();
      }
      
      return backup;
   }
   
   /**
    * Reports if the current block has the given property.
    *
    * @param    prop
    *           The property bit flag to test.  Should be one of the constants
    *           <code>PROP_ERROR, PROP_ENDKEY, PROP_STOP or PROP_QUIT</code>.
    *
    * @return   <code>true</code> if the current block has the given 
    *           property bit set in its properties bitmap.
    */
   private static boolean hasProperty(int prop)
   {
      BlockDefinition blk = workArea.obtain().blocks.peek();
      return (blk.props & prop) != 0;
   }
   
   /**
    * Write a formatted log entry with the error text and some state based 
    * on the current block.
    *
    * @param    lvl
    *           The logging level.
    * @param    location
    *           The event or call site from which the log entry was
    *           generated.
    * @param    txt
    *           The text to insert into the log.  May be <code>null</code>.
    */
   private static void log(Level           lvl,
                           String          location,
                           String          txt)
   {
      log(lvl, location, txt, null);
   }
      
   /**
    * Write a formatted log entry with the error text and some state based 
    * on the current block.
    *
    * @param    lvl
    *           The logging level.
    * @param    location
    *           The event or call site from which the log entry was
    *           generated.
    * @param    txt
    *           The text to insert into the log.  May be <code>null</code>.
    * @param    trw
    *           The throwable that caused this logging to occur or
    *           <code>null</code> if no exception is present.
    */
   private static void log(Level     lvl,
                           String    location,
                           String    txt,
                           Throwable trw)
   {
      if (LOG.isLoggable(lvl))
      {
         WorkArea        wa  = workArea.obtain();
         BlockDefinition blk = null;
         
         if (wa.depth() > 0)
         {
            blk = wa.blocks.peek();
         }
         
         String err = LogHelper.generate("%s %s %s", 
                                         wa.toString(),
                                         (blk == null ? "" : blk.toString()),
                                         (txt == null ? "" : txt));
                                         
         if (trw == null)
         {
            LOG.logp(lvl, "TransactionManager." + location, "", err);
         }
         else
         {
            LOG.logp(lvl, "TransactionManager." + location, "", err, trw);
         }
      }
   }
   
   /**
    * Stores global data relating to the state of the current context's
    * transactions.  Includes convenience methods for notifying listeners.
    */
   private static class WorkArea
   {  
      /** Sprintf format specification for string representation. */
      public static final String spec = "<depth = %d; trans_level = %d; "   +
                                        "trans_label = %s; "                +
                                        "rollback_scope = %d; "             +
                                        "rollback_label = %s; "             +
                                        "rollback_pending = %b; in_quit = " +
                                        "%b; retry_scope = %d; "            +
                                        "retry_label = %s; "                +
                                        "ignore_err = %b>";
                                        
      /** Registry of all objects needing start and end scope callbacks. */
      private final List<Scopeable> scopeList = new ArrayList<Scopeable>();
      
      /**
       * Map of Commitables to get a commit notification after all other
       * commits occur in a full transaction;  keys are the listeners, values
       * are flags indicating whether listeners are auto-removed.
       */
      private final Map<Object, Boolean> masterCommit = new LinkedHashMap<Object, Boolean>();
      
      /**
       * Map of Finalizables to get a finish notification after all other
       * finish notifications occur in a full transaction;  keys are the
       * listeners, values are flags indicating whether listeners are
       * auto-removed.
       */
      private final Map<Object, Boolean> masterFinish = new LinkedHashMap<Object, Boolean>();
      
      /** Map of BatchListeners to get batch start and stop notifications. */
      private final List<BatchListener> batchListeners = new LinkedList<BatchListener>();
      
      /** Registry of all objects needing undo support. */
      private PartitionedArray<Undoable> undoList =
         new PartitionedArray<Undoable>();
      
      /** Stack level of the full transaction if one is currently active. */ 
      private int transLevel = -1;
      
      /** Defines the scope at which rollback will occur. */ 
      private int rollbackScope = -1;      
      
      /** Flag to denote if a rollback is pending. */ 
      private boolean rollbackPending = false;      
      
      /** Flag to denote if the exit is due to a quit condition. */
      private boolean processingQuit = false;
      
      /** Defines the scope at which a pending retry will occur. */ 
      private int retryScope = -1;      
      
      /**
       * Defines if an error condition should be ignored until reaching the 
       * caller of a top-level block.
       */
      private boolean ignoreError = false;
      
      /** Throwable caught during commit/rollback processing but deferred. */
      private Throwable deferredError = null;
      
      /** Enables/disables nested error support at runtime. */
      private boolean nested = false;
      
      /** Exception caught during processing another error. */
      private ConditionException nestedError = null;
      
      /** Defines a single global scope which is not kept in the stack. */ 
      private BlockDefinition globalBlock = null;      
      
      /**
       * Tracks block definitions with the top of the stack as the innermost
       * scope.
       */
      private Stack<BlockDefinition> blocks = new Stack<BlockDefinition>();
      
      /** <code>Finalizable</code>s which need deferred registration. */
      private List<Finalizable> deferredExtFinalizables = null;
      
      /** <code>Undoable</code>s which need deferred registration. */
      private List<Undoable> deferredUndoables = null;

      /** Collection or map currently being iterated. */
      private Object nowIterating = null;
      
      /** Current operation. */
      private int op = OP_NONE;
      
      /** Registered STOP condition handlers */
      private List<StopConditionVetoHandler> stopHandlers =
         new LinkedList<StopConditionVetoHandler>();

      /**
        * A stack of registered query off-end listeners, which interrupt the
        * FOR or FOR EACH block when a {@link QueryOffEndException} is 
        * caught (the user defined END-KEY action is ignored and will perform
        * a UNDO, LEAVE).
        */
      private ArrayDeque<Set<QueryOffEndListener>> offEndListeners = 
         new ArrayDeque<Set<QueryOffEndListener>>(10);

      /**
       * Registered <code>Resettable</code>s which should be reset at
       * specific points of block processing.
       */
      private List<Resettable> resettables = new LinkedList<Resettable>();

      /**
       * A set of queries registered during FOR or FOR EACH block 
       * initialization. Only {@link QueryOffEndException} triggered by these
       * queries can end a FOR or FOR EACH block, even if the END-KEY action 
       * is overridden.
       */
      private Set<P2JQuery> offEndQueries = new HashSet<P2JQuery>(3);

      /**
       * Flag indicating that the block executes its {@link Block#init()} 
       * method. Only variants of FOR or FOR EACH blocks will set the flag to
       * <code>true</code>. During block initialization, all created queries
       * will be collected in the {@link #offEndQueries} set. 
       */
      private boolean offEndRegistration = false;

      /**
       * Flag indicating the processing should be terminated.
       */
      private volatile boolean abortProcessing = false;
      
      /** Flag that disables calls to the client side. */
      private boolean headless = false;
      
      /**
       * Report scope depth.
       *
       * @return  Number of blocks currently on stack.
       */
      private int depth()
      {
         return blocks.size();
      }
      
      /**
       * Notify all master Finalizables of a retry event.  Listeners are
       * notified in the order in which they were registered.
       */
      private void notifyMasterRetry()
      {
         if (LOG.isLoggable(Level.FINE))
         {
            log(Level.FINE, "notifyMasterRetry", null);
         }
         
         try
         {
            nowIterating = masterFinish;
            op = OP_RETRY;
            
            Iterator<Object> iter = masterFinish.keySet().iterator();
            while (iter.hasNext())
            {
               // Process notification, deferring any error until we are in a
               // safe state to handle it.
               Finalizable target = (Finalizable) iter.next();
               try
               {
                  target.retry();
               }
               catch (Throwable thr)
               {
                  deferError("master retry", thr);
               }
            }
         }
         
         finally
         {
            nowIterating = null;
            op = OP_NONE;
         }
      }
      
      /**
       * Notify all master Finalizables of an iterate or finished event.
       * After each notification, the corresponding Finalizable is removed,
       * if it was registered as auto-removable.  Listeners are notified in
       * the order in which they were registered.
       *
       * @param   iterate
       *          <code>true</code> to notify of an iterate event;
       *          <code>false</code> to notify of a finished event.
       */
      private void notifyMasterFinish(boolean iterate)
      {
         if (LOG.isLoggable(Level.FINE))
         {
            log(Level.FINE,
                "notifyMasterFinish",
                "iterate = " + Boolean.toString(iterate));
         }
         
         try
         {
            nowIterating = masterFinish;
            op = iterate ? OP_ITERATE : OP_FINISHED;
            
            Iterator<Object> iter = masterFinish.keySet().iterator();
            while (iter.hasNext())
            {
               // Process notification, deferring any error until we are in a
               // safe state to handle it.
               Finalizable target = (Finalizable) iter.next();
               if (iterate)
               {
                  try
                  {
                     target.iterate();
                  }
                  catch (Throwable thr)
                  {
                     deferError("master iterate", thr);
                  }
               }
               else
               {
                  try
                  {
                     target.finished();
                  }
                  catch (Throwable thr)
                  {
                     deferError("master finish", thr);
                  }
               }
            }
         }
         
         finally
         {
            nowIterating = null;
            op = OP_NONE;
         }
      }
      
      /**
       * Notify all master Commitables of a rollback or commit event.
       * After each notification, the corresponding Commitable is removed,
       * if it was registered as auto-removable and <code>autoRemove</code>
       * is <code>true</code>.  Listeners are notified in the order in which
       * they were registered.
       *
       * @param   rollback
       *          <code>true</code> to notify of a rollback event;
       *          <code>false</code> to notify of a commit event.
       */
      private void notifyMasterCommit(boolean rollback)
      {
         if (LOG.isLoggable(Level.FINE))
         {
            log(Level.FINE,
                "notifyMasterCommit",
                "rollback = " + Boolean.toString(rollback));
         }
         
         try
         {
            nowIterating = masterCommit;
            
            Iterator<Object> iter = masterCommit.keySet().iterator();
            while (iter.hasNext())
            {
               // Process notification, deferring any error until we are in a
               // safe state to handle it.
               Commitable target = (Commitable) iter.next();
               if (rollback)
               {
                  try
                  {
                     target.rollback(true);
                  }
                  catch (Throwable thr)
                  {
                     deferError("master rollback", thr);
                  }
               }
               else
               {
                  try
                  {
                     target.commit(true);
                  }
                  catch (Throwable thr)
                  {
                     deferError("master commit", thr);
                  }
               }
            }
         }
         
         finally
         {
            nowIterating = null;
         }
      }
      
      /**
       * Iterate through registered, master Finalizables and deregister those
       * marked for auto-remove.
       */
      private void autoRemoveMasterFinish()
      {
         if (nowIterating == masterFinish)
         {
            // prevent ConcurrentModificationException
            if (LOG.isLoggable(Level.WARNING))
            {
               log(Level.WARNING,
                   "autoRemoveMasterFinish",
                   "Illegal attempt to auto-remove master Finalizables",
                   new Throwable());
            }
            
            return;
         }
         
         autoRemove(masterFinish);
      }
      
      /**
       * Iterate through registered, master Commitables and deregister those
       * marked for auto-remove.
       */
      private void autoRemoveMasterCommit()
      {
         if (nowIterating == masterCommit)
         {
            // prevent ConcurrentModificationException
            if (LOG.isLoggable(Level.WARNING))
            {
               log(Level.WARNING,
                   "autoRemoveMasterCommit",
                   "Illegal attempt to auto-remove master Commitables",
                   new Throwable());
            }
            
            return;
         }
         
         autoRemove(masterCommit);
      }
      
      /**
       * Iterate through the given map entries and remove all those whose
       * values are <code>true</code>.  The key is a <code>Commitable</code>
       * or <code>Finalizable</code> target, the value is its registered
       * auto-remove setting.
       *
       * @param   map
       *          Map of registered targets and their auto-remove settings.
       */
      private void autoRemove(Map<Object, Boolean> map)
      {
         Iterator<Map.Entry<Object, Boolean>> iter = map.entrySet().iterator();
         
         while (iter.hasNext())
         {
            if (iter.next().getValue().booleanValue())
            {
               iter.remove();
            }
         }
      }
      
      /**
       * Render this work area's data as a human readable string.
       *
       * @return   The human readable form of the global data.
       */
      public String toString()
      {
         BlockDefinition tblk = null;
         BlockDefinition rollbackBlk = null;
         BlockDefinition retryBlk = null;
         int size = blocks.size();
         
         if (transLevel != -1 && transLevel < size)
         {
            tblk = blocks.elementAt(transLevel);
         }
         
         if (rollbackScope != -1 && rollbackScope < size)
         {
            rollbackBlk = blocks.elementAt(rollbackScope);
         }
         
         if (retryScope != -1 && retryScope < size)
         {
            retryBlk = blocks.elementAt(retryScope);
         }
         
         Object[] parms = new Object[]
         {
            depth(),
            transLevel,
            (tblk == null) ? null : tblk.label,
            rollbackScope,
            (rollbackBlk == null) ? null : rollbackBlk.label,
            rollbackPending,
            processingQuit,
            retryScope,
            (retryBlk == null) ? null : retryBlk.label,
            ignoreError
         };
         
         return String.format(spec, parms);
      }  
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Get the weight of this context-local var.
       * 
       * @return   Always {@link WeightFactor#LEVEL_4}.
       */
      @Override
      public WeightFactor getWeight()
      {
         return WeightFactor.LEVEL_4;
      };

      /**
       * Obtains the context-local instance of the contained global work
       * area.  The first time this method is invoked in a context, a new
       * <code>WorkArea</code> instance is lazily created.  This is done here
       * rather than using the more elegant mechanism of overriding the
       * superclass' <code>initialValue</code> method, because we must ensure
       * we do not trigger an infinite loop.  Infinite looping can occur
       * because a new work area requires that each {@link ScopeableFactory}
       * create its {@link Scopeable} instance.  These objects may call back
       * into the <code>TransactionManager</code> to register for various
       * service hooks, which will in turn trigger a call to this method.
       * By implementing the <code>WorkArea</code> instantiation here, we can
       * set the <code>ContextContainer</code>'s contents <i>before</i> the
       * factories are asked to create their <code>Scopeable</code>s.  Thus,
       * any recursion into this method during <code>Scopeable</code>
       * initialization is safe.
       *
       * @return  The work area associated with this context.
       */
      public WorkArea obtain() 
      {
         WorkArea wa = get();
         if (wa == null)
         {
            wa = new WorkArea();
            wa.globalBlock = new BlockDefinition("globalScope",
                                                 NO_TRANSACTION,
                                                 PROP_NONE,
                                                 false,
                                                 false,
                                                 false,
                                                 false,
                                                 false,
                                                 BlockType.UNKNOWN,
                                                 null);
            
            
            // detect if headless mode is needed
            wa.headless = _isHeadless();
            
            // the following call MUST occur before creating Scopeables
            // below, in case a Scopeable initializer calls back into the
            // TransactionManager in such a way as to call this class'
            // obtain() method (prevents infinite loop and stack overflow)
            set(wa);
            
            // create all scopeable objects to be registered using the
            // registered Scopeable factories
            Iterator<ScopeableFactory> iter = scopeableFactories.iterator();
            
            while (iter.hasNext())
            {
               ScopeableFactory factory = iter.next();
               Scopeable        scope   = factory.createScopeable();
               
               // non-null return values must be added to the list otherwise
               // it is silently bypassed
               if (scope != null)
                  wa.scopeList.add(scope);
            }
         }
         
         return wa;
      }   
      
      /**
       * Return <code>null</code> rather than providing a true initial value;
       * initial value is instead lazily created in {@link #obtain}.
       *
       * @return  <code>null</code>.
       */
      protected final WorkArea initialValue()
      {
         return null;
      }   
   }
}
